import asyncio
import json
import os
from dataclasses import dataclass
from typing import AsyncGenerator
from urllib.error import URLError
from urllib.request import urlopen

import aceai
from msgspec import Struct
from opentelemetry.context import Context

from aceai.agent.citations import TurnCitation, message_with_citations
from aceai.agent.config import (
    AgentAppConfig,
    ReasoningLevel,
    replace_config,
    save_config,
)
from aceai.agent.features.delegation import (
    ChildAgentResult,
    build_child_agent_result,
    build_restored_delegated_child_agent,
    reset_delegated_child_run_recorder,
    set_delegated_child_run_recorder,
)
from aceai.agent.memory.ideas import Idea, IdeaStore
from aceai.agent.project import ProjectMetadata, default_project
from aceai.agent.provider_auth import default_api_key_for_provider
from aceai.agent.provider_catalog import (
    api_key_env,
    context_window_for_model_any_provider,
    model_options,
    supported_models,
    supports_reasoning_effort,
    supports_reasoning_effort_any_provider,
)
from aceai.agent.session import (
    EventLog,
    MAIN_THREAD_ID,
    SessionEvent,
    SessionRecorder,
    SessionState,
    SessionStore,
)
from aceai.agent.session_service import AgentSessionSnapshot, SessionService
from aceai.agent.memory.subagent_artifacts import SubagentArtifactStore
from aceai.core.helpers.string import uuid_str
from aceai.core import (
    Agent,
    AgentRunContext,
    Executor,
    ToolApprovalDecision,
    ToolExecutionError,
)
from aceai.core.events import (
    AgentEvent,
    ContextCompactionFailedEvent,
    ContextCompactionStartedEvent,
    ContextCompressedEvent,
    LLMCompletedEvent,
    RunFailedEvent,
    RunCompletedEvent,
    RunSuspendedEvent,
    ToolCompletedEvent,
    ToolFailedEvent,
)
from aceai.core.models import AgentStep
from aceai.core.models import ToolApprovalRequest
from aceai.llm.models import LLMMessage, LLMRequestMeta, LLMResponse

PYPI_PROJECT_JSON_URL = "https://pypi.org/pypi/aceai/json"


class UpdateCheckResult(Struct, frozen=True, kw_only=True):
    current_version: str
    latest_version: str

    @property
    def has_update(self) -> bool:
        return _version_parts(self.latest_version) > _version_parts(
            self.current_version
        )


class AgentAppEvent(Struct, frozen=True, kw_only=True):
    thread_id: str
    agent_id: str
    event: AgentEvent


class AppRuntimeInfo(Struct, frozen=True, kw_only=True):
    provider_name: str
    selected_model: str
    default_model: str
    reasoning_level: ReasoningLevel
    supports_reasoning: bool
    context_window: int | None
    max_steps: str


@dataclass
class ChildThreadRuntime:
    thread_id: str
    agent_id: str
    run: AgentRunContext
    handoff: asyncio.Future[ChildAgentResult]
    task: asyncio.Task[None]
    intervention_count: int = 0


class AceAgentApp:
    """Reusable agent app runtime used by UI and future ports."""

    def __init__(
        self,
        agent: Agent,
        *,
        provider_name: str,
        selected_model: str,
        reasoning_level: ReasoningLevel = "auto",
        initial_history: list[LLMMessage] | None = None,
        session_service: SessionService | None = None,
        session_store: SessionStore | None = None,
        session_recorder: SessionRecorder | None = None,
        session_id: str | None = None,
        idea_store: IdeaStore | None = None,
        project: ProjectMetadata | None = None,
        trace_ctx: Context | None = None,
    ) -> None:
        self._agent = agent
        self._provider_name = provider_name
        self._selected_model = selected_model
        self._project = project or (
            session_store.project
            if session_store is not None
            else default_project()
        )
        self._reasoning_level: ReasoningLevel = "auto"
        self._request_meta: LLMRequestMeta = {"model": selected_model}
        self._set_reasoning_level_internal(reasoning_level)
        self._session_service = session_service or SessionService(
            store=session_store,
            recorder=session_recorder,
            session_id=session_id,
            project=self._project,
        )
        self._subagent_artifacts = SubagentArtifactStore(self._session_service.store.root)
        self._trace_ctx = trace_ctx
        snapshot = None
        if initial_history is None and self._session_service.session_id is not None:
            snapshot = self._session_service.snapshot(self._session_service.session_id)
            self._llm_history = list(snapshot.history)
        else:
            self._llm_history = list(initial_history or [])
        self._active_run: AgentRunContext | None = None
        self._thread_runs: dict[str, AgentRunContext] = {}
        self._last_context_event_id: str | None = None
        self._last_context_event_ids: dict[str, str] = {}
        self._active_thread_id = MAIN_THREAD_ID
        self._thread_agents: dict[str, Agent] = {MAIN_THREAD_ID: self._agent}
        self._child_runtimes: dict[str, ChildThreadRuntime] = {}
        self._steered_child_run_ids: set[str] = set()
        self._stale_child_runtime_tasks: set[asyncio.Task[None]] = set()
        self._queued_questions: dict[str, list[str]] = {}
        self._app_event_subscribers: set[asyncio.Queue[AgentAppEvent | None]] = set()
        self._idea_store = idea_store or IdeaStore()
        self._approved_tool_names_by_thread: dict[str, set[str]] = {}
        self._child_context_event_ids: dict[str, str] = {}
        self._update_check_completed = False
        self._update_check_lock = asyncio.Lock()
        session_id = self._session_service.session_id
        if session_id is not None:
            if snapshot is None:
                snapshot = self._session_service.snapshot(session_id)
            self._approved_tool_names_by_thread[MAIN_THREAD_ID] = (
                _approved_tool_names_from_event_log(snapshot.event_log)
            )
            self._last_context_event_id = _last_context_source_event_id(snapshot.event_log)
            if self._last_context_event_id is not None:
                self._last_context_event_ids[MAIN_THREAD_ID] = self._last_context_event_id

    @property
    def agent(self) -> Agent:
        return self._agent

    @property
    def provider_name(self) -> str:
        return self._provider_name

    @property
    def selected_model(self) -> str:
        return self._selected_model

    @property
    def reasoning_level(self) -> ReasoningLevel:
        return self._reasoning_level

    @property
    def model_options_text(self) -> str:
        return model_options_text_for(self._provider_name)

    def is_model_supported(self, model: str) -> bool:
        return model in supported_models(self._provider_name)

    def runtime_info(self) -> AppRuntimeInfo:
        agent = self._agent
        return AppRuntimeInfo(
            provider_name=self._provider_name,
            selected_model=self._selected_model,
            default_model=str(agent.default_model),
            reasoning_level=self._reasoning_level,
            supports_reasoning=supports_reasoning_for_model(self._selected_model),
            context_window=context_window_for_model(self._selected_model),
            max_steps=f"{agent.max_steps}",
        )

    def skill_summary_lines(self) -> list[str]:
        return [
            f"{skill.name}: {skill.description} ({skill.skill_file})"
            for skill in self._agent.skill_registry.get_skills()
        ]

    def tool_summary_lines(self) -> list[str]:
        executor = self._agent.executor
        if not isinstance(executor, Executor):
            return []
        lines: list[str] = []
        for tool in executor.tools.values():
            tags = ", ".join(tool.metadata.tags)
            tag_text = f" [{tags}]" if tags else ""
            lines.append(f"{tool.name}{tag_text}: {tool.description}")
        return lines

    def hosted_tool_summary_lines(self) -> list[str]:
        return [
            f"{tool.provider_name}:{tool.native_name}"
            for tool in self._agent.hosted_tools
        ]

    @property
    def session_service(self) -> SessionService:
        return self._session_service

    @property
    def project(self) -> ProjectMetadata:
        return self._project

    @property
    def project_name(self) -> str:
        return self._project.name

    @property
    def session_id(self) -> str | None:
        return self._session_service.session_id

    @property
    def session_recorder(self) -> SessionRecorder | None:
        return self._session_service.recorder

    @property
    def llm_history(self) -> list[LLMMessage]:
        return list(self._llm_history)

    @property
    def active_run(self) -> AgentRunContext | None:
        return self._active_run

    @property
    def queued_questions(self) -> tuple[str, ...]:
        return tuple(self._queued_questions_for_active_thread())

    @property
    def active_thread_accepts_user_turn(self) -> bool:
        return self._active_thread_id not in self._child_runtimes

    @property
    def is_running_suspended(self) -> bool:
        run = self._active_run
        return run is not None and run.status == "suspended"

    def ensure_session(self) -> str:
        return self._session_service.ensure_session()

    def switch_session(self, session_id: str) -> AgentSessionSnapshot:
        snapshot = self._session_service.attach_session(session_id)
        self._llm_history = list(snapshot.history)
        self._active_run = None
        self._thread_runs = {}
        self._active_thread_id = MAIN_THREAD_ID
        self._thread_agents = {MAIN_THREAD_ID: self._agent}
        self._child_runtimes = {}
        self._queued_questions = {}
        self._approved_tool_names_by_thread = {
            MAIN_THREAD_ID: _approved_tool_names_from_event_log(snapshot.event_log)
        }
        self._last_context_event_id = _last_context_source_event_id(snapshot.event_log)
        self._last_context_event_ids = {}
        if self._last_context_event_id is not None:
            self._last_context_event_ids[MAIN_THREAD_ID] = self._last_context_event_id
        return snapshot

    @property
    def active_thread_id(self) -> str:
        return self._active_thread_id

    def switch_thread(self, thread_id: str) -> AgentSessionSnapshot:
        session_id = self.session_id
        if session_id is None:
            raise RuntimeError("AceAI session is not active")
        snapshot = self._session_service.snapshot_thread(session_id, thread_id)
        self._active_thread_id = thread_id
        if thread_id == MAIN_THREAD_ID:
            self._llm_history = list(snapshot.history)
        self._approved_tool_names_by_thread[thread_id] = (
            _approved_tool_names_from_event_log(snapshot.event_log)
        )
        self._active_run = self._thread_runs.get(thread_id)
        latest_context_event_id = _last_context_source_event_id(snapshot.event_log)
        if latest_context_event_id is not None:
            self._last_context_event_ids[thread_id] = latest_context_event_id
        return snapshot

    def restore_history_from_active_session(self) -> None:
        session_id = self.session_id
        if session_id is None:
            self._llm_history = []
            self._active_run = None
            self._thread_runs = {}
            self._queued_questions = {}
            self._approved_tool_names_by_thread = {}
            self._last_context_event_id = None
            self._last_context_event_ids = {}
            return
        snapshot = self._session_service.snapshot_thread(
            session_id,
            self._active_thread_id,
        )
        if self._active_thread_id == MAIN_THREAD_ID:
            self._llm_history = snapshot.history
        latest_context_event_id = _last_context_source_event_id(snapshot.event_log)
        self._last_context_event_id = (
            latest_context_event_id if self._active_thread_id == MAIN_THREAD_ID else None
        )
        if latest_context_event_id is not None:
            self._last_context_event_ids[self._active_thread_id] = latest_context_event_id
        self._active_run = self._thread_runs.get(self._active_thread_id)

    def cancel_active_turn(self) -> None:
        self._active_run = None
        self._thread_runs.pop(self._active_thread_id, None)

    def enqueue_turn(self, question: str) -> int:
        if not self.active_thread_accepts_user_turn:
            raise RuntimeError(
                "Delegated subagent thread is still running; switch to main or wait "
                "for delegate_to_subagent to finish before sending a new message."
            )
        questions = self._queued_questions_for_active_thread()
        questions.append(question)
        return len(questions)

    def steer_active_child_thread(self, question: str) -> bool:
        thread_id = self._active_thread_id
        runtime = self._child_runtimes.get(thread_id)
        if runtime is None:
            return False
        if question == "":
            raise ValueError("child steer question cannot be empty")
        if runtime.task.done():
            self._child_runtimes.pop(thread_id, None)
            return False
        child_agent = self._agent_for_thread(thread_id)
        session_id = self.session_id
        if session_id is None:
            raise RuntimeError("AceAI session is not active")
        event_log = self._session_service.store.load_thread_event_log(
            session_id,
            thread_id,
        )
        child_run = child_agent.create_resume_run(
            question,
            event_log.replay_llm_history(),
            trace_ctx=self._trace_ctx,
            **self._request_meta_for_run(),
        )
        child_run.run_state.tools.approved_tool_names = (
            self._approved_tool_names_for_thread(thread_id)
        )
        runtime.intervention_count += 1
        self._steered_child_run_ids.add(runtime.run.run_id)
        runtime.run = child_run
        self._thread_runs[thread_id] = child_run
        self._active_run = child_run
        event_id = self._record_child_steer_event(
            thread_id=thread_id,
            agent_id=runtime.agent_id,
            run_id=child_run.run_id,
            question=question,
        )
        self._child_context_event_ids[thread_id] = event_id
        self._last_context_event_ids[thread_id] = event_id
        old_task = runtime.task
        self._stale_child_runtime_tasks.add(old_task)
        old_task.cancel()
        runtime.task = asyncio.create_task(
            self._run_child_thread_runtime(
                thread_id=thread_id,
                child_agent=child_agent,
                child_run=child_run,
                handoff=runtime.handoff,
            )
        )
        runtime.task.add_done_callback(
            lambda completed_task: self._complete_child_runtime_task(
                thread_id=thread_id,
                task=completed_task,
                handoff=runtime.handoff,
            )
        )
        return True

    def pop_queued_turn(self) -> str | None:
        questions = self._queued_questions_for_active_thread()
        if not questions:
            return None
        return questions.pop(0)

    def take_queued_turn(self, index: int) -> str:
        return self._queued_questions_for_active_thread().pop(index)

    def cancel_queued_turn(self, index: int) -> str:
        return self._queued_questions_for_active_thread().pop(index)

    def switch_model(
        self,
        model: str,
        *,
        reasoning_level: ReasoningLevel | None = None,
    ) -> None:
        if not self.is_model_supported(model):
            raise ValueError(f"Unsupported model: {model}")
        self._selected_model = model
        self._request_meta["model"] = model
        next_level = (
            reasoning_level if reasoning_level is not None else self._reasoning_level
        )
        self._set_reasoning_level_internal(next_level)
        self.persist_session_state()

    def set_reasoning_level(self, level: ReasoningLevel) -> None:
        self._set_reasoning_level_internal(level)
        self.persist_session_state()

    def _set_reasoning_level_internal(self, level: ReasoningLevel) -> None:
        if level != "auto" and not supports_reasoning_effort(
            self._provider_name,
            self._selected_model,
        ):
            level = "auto"
        self._reasoning_level = level
        if level == "auto":
            self._request_meta.pop("reasoning", None)
        else:
            self._request_meta["reasoning"] = {"effort": level, "summary": "auto"}

    def persist_session_state(self) -> None:
        session_id = self.session_id
        if session_id is None:
            return
        self._session_service.update_state(
            session_id,
            SessionState(
                selected_provider=self._provider_name,
                selected_model=self._selected_model,
            ),
        )

    def capture_idea(self, content: str) -> Idea:
        return self._idea_store.capture(
            content,
            project=self._project,
            source_session_id=self.session_id,
        )

    def list_ideas(self) -> list[Idea]:
        return self._idea_store.list_for_display(current_project=self._project)

    def delete_idea(self, index: int) -> Idea:
        return self._idea_store.delete_displayed(
            index,
            current_project=self._project,
        )

    def update_idea(self, index: int, content: str) -> Idea:
        return self._idea_store.update_displayed(
            index,
            content,
            current_project=self._project,
        )

    def pending_approval_request(
        self,
        *,
        thread_id: str | None = None,
    ) -> ToolApprovalRequest | None:
        run = self._run_for_thread(thread_id or self._active_thread_id)
        if run is None:
            return None
        pending = run.run_state.pending_approval
        if pending is None:
            return None
        return pending.request

    async def check_for_updates(self) -> UpdateCheckResult | None:
        async with self._update_check_lock:
            if self._update_check_completed:
                return None
            self._update_check_completed = True
            result = await check_for_updates()
            if result is None or not result.has_update:
                return None
            return result

    async def start_turn(
        self,
        question: str,
        *,
        citations: tuple[TurnCitation, ...] = (),
    ) -> AsyncGenerator[AgentEvent, None]:
        thread_id = self._active_thread_id
        async for app_event in self.start_turn_events(
            question,
            citations=citations,
        ):
            if app_event.thread_id == thread_id:
                yield app_event.event

    async def start_turn_events(
        self,
        question: str,
        *,
        citations: tuple[TurnCitation, ...] = (),
    ) -> AsyncGenerator[AgentAppEvent, None]:
        self.ensure_session()
        self.persist_session_state()
        thread_id = self._active_thread_id
        if thread_id in self._child_runtimes:
            self.steer_active_child_thread(question)
            return
        agent = self._agent_for_thread(thread_id)
        history = self._history_for_thread(thread_id)
        llm_question = message_with_citations(question, citations)
        self._active_run = agent.create_resume_run(
            llm_question,
            history,
            trace_ctx=self._trace_ctx,
            **self._request_meta_for_run(),
        )
        self._thread_runs[thread_id] = self._active_run
        self._active_run.run_state.tools.approved_tool_names = (
            self._approved_tool_names_for_thread(thread_id)
        )
        self._last_context_event_id = self._session_service.record_user_message(
            question,
            citations=citations,
            run_id=self._active_run.run_id,
            thread_id=thread_id,
            agent_id="" if thread_id == MAIN_THREAD_ID else agent.agent_id,
        )
        self._last_context_event_ids[thread_id] = self._last_context_event_id
        async for event in self._consume_run_stream_events(
            self._active_run,
            agent.execute(self._active_run),
            thread_id=thread_id,
            agent_id="" if thread_id == MAIN_THREAD_ID else agent.agent_id,
        ):
            yield event

    async def approve_tool(
        self,
        *,
        thread_id: str | None = None,
        run_id: str | None = None,
        tool_call_id: str | None = None,
    ) -> AsyncGenerator[AgentEvent, None]:
        target_thread_id = thread_id or self._active_thread_id
        request = self.pending_approval_request(thread_id=target_thread_id)
        if request is None:
            raise RuntimeError("No pending tool approval")
        self._validate_approval_target(
            thread_id=target_thread_id,
            request=request,
            run_id=run_id,
            tool_call_id=tool_call_id,
        )
        async for event in self._resume_approval(
            ToolApprovalDecision.approve(request),
            thread_id=target_thread_id,
        ):
            yield event

    async def reject_tool(
        self,
        reason: str,
        *,
        thread_id: str | None = None,
        run_id: str | None = None,
        tool_call_id: str | None = None,
    ) -> AsyncGenerator[AgentEvent, None]:
        target_thread_id = thread_id or self._active_thread_id
        request = self.pending_approval_request(thread_id=target_thread_id)
        if request is None:
            raise RuntimeError("No pending tool approval")
        self._validate_approval_target(
            thread_id=target_thread_id,
            request=request,
            run_id=run_id,
            tool_call_id=tool_call_id,
        )
        if reason == "":
            reason = "rejected by caller"
        async for event in self._resume_approval(
            ToolApprovalDecision.reject(request, reason=reason),
            thread_id=target_thread_id,
        ):
            yield event

    async def _resume_approval(
        self,
        decision: ToolApprovalDecision,
        *,
        thread_id: str,
    ) -> AsyncGenerator[AgentEvent, None]:
        run = self._current_run(thread_id=thread_id)
        async for event in self._consume_run_stream(
            run,
            self._agent_for_thread(thread_id).resume_approval(run, decision),
            thread_id=thread_id,
            agent_id="" if thread_id == MAIN_THREAD_ID else run.agent_id,
        ):
            yield event

    async def _consume_run_stream(
        self,
        run: AgentRunContext,
        stream: AsyncGenerator[AgentEvent, None],
        *,
        thread_id: str,
        agent_id: str,
    ) -> AsyncGenerator[AgentEvent, None]:
        async for app_event in self._consume_run_stream_events(
            run,
            stream,
            thread_id=thread_id,
            agent_id=agent_id,
        ):
            if app_event.thread_id == thread_id:
                yield app_event.event

    async def _consume_run_stream_events(
        self,
        run: AgentRunContext,
        stream: AsyncGenerator[AgentEvent, None],
        *,
        thread_id: str,
        agent_id: str,
    ) -> AsyncGenerator[AgentAppEvent, None]:
        queue: asyncio.Queue[AgentAppEvent | None] = asyncio.Queue()
        self._app_event_subscribers.add(queue)

        async def produce_events() -> None:
            recorder_token = set_delegated_child_run_recorder(self)
            try:
                async for event in stream:
                    if isinstance(event, ContextCompressedEvent):
                        self._record_context_checkpoint(event, thread_id=thread_id)
                    event = self._archive_subagent_artifact(event)
                    persisted_event_id: str | None = None
                    if not isinstance(
                        event,
                        ContextCompactionFailedEvent
                        | ContextCompactionStartedEvent
                        | ContextCompressedEvent,
                    ):
                        persisted_event_id = self._session_service.record_agent_event(
                            event,
                            thread_id=thread_id,
                            agent_id=agent_id,
                        )
                    if persisted_event_id is not None and _agent_event_updates_context(
                        event
                    ):
                        self._last_context_event_ids[thread_id] = persisted_event_id
                        if thread_id == MAIN_THREAD_ID:
                            self._last_context_event_id = persisted_event_id
                    if isinstance(event, RunCompletedEvent):
                        self._finish_run_turn(
                            run,
                            event.final_answer,
                            thread_id=thread_id,
                        )
                    self._emit_app_event(
                        thread_id=thread_id,
                        agent_id=agent_id,
                        event=event,
                    )
            finally:
                reset_delegated_child_run_recorder(recorder_token)
                await stream.aclose()
                queue.put_nowait(None)

        producer = asyncio.create_task(produce_events())
        try:
            while True:
                item = await queue.get()
                if item is None:
                    break
                yield item
        finally:
            self._app_event_subscribers.discard(queue)
            if not producer.done():
                producer.cancel()
            await producer

    def _run_for_thread(self, thread_id: str) -> AgentRunContext | None:
        return self._thread_runs.get(thread_id)

    def _approved_tool_names_for_thread(self, thread_id: str) -> set[str]:
        if thread_id not in self._approved_tool_names_by_thread:
            session_id = self.session_id
            if session_id is None:
                self._approved_tool_names_by_thread[thread_id] = set()
            else:
                snapshot = self._session_service.snapshot_thread(session_id, thread_id)
                self._approved_tool_names_by_thread[thread_id] = (
                    _approved_tool_names_from_event_log(snapshot.event_log)
                )
        return self._approved_tool_names_by_thread[thread_id]

    def _current_run(self, *, thread_id: str | None = None) -> AgentRunContext:
        run = self._run_for_thread(thread_id or self._active_thread_id)
        if run is None:
            raise RuntimeError("AceAI run is not active")
        return run

    def _validate_approval_target(
        self,
        *,
        thread_id: str,
        request: ToolApprovalRequest,
        run_id: str | None,
        tool_call_id: str | None,
    ) -> None:
        run = self._current_run(thread_id=thread_id)
        if run_id is not None and run.run_id != run_id:
            raise RuntimeError("approval target run_id does not match pending run")
        if tool_call_id is not None and request.call.call_id != tool_call_id:
            raise RuntimeError("approval target tool_call_id does not match pending tool call")

    def _finish_run_turn(
        self,
        run: AgentRunContext,
        answer: str,
        *,
        thread_id: str,
    ) -> None:
        if thread_id != MAIN_THREAD_ID:
            return
        self._llm_history = list(run.context.context[1:])
        self._llm_history.append(LLMMessage.build(role="assistant", content=answer))

    def _record_context_checkpoint(
        self,
        event: ContextCompressedEvent,
        *,
        thread_id: str,
    ) -> None:
        included_event_id = self._last_context_event_ids.get(thread_id)
        if included_event_id is None:
            raise RuntimeError("Context checkpoint has no included transcript event")
        if thread_id == MAIN_THREAD_ID:
            self._session_service.record_context_checkpoint(
                event,
                included_event_id=included_event_id,
            )
            return
        session_id = self.session_id
        if session_id is None:
            raise RuntimeError("AceAI session is not active")
        self._session_service.context_checkpoint_store.record_checkpoint(
            session_id=session_id,
            thread_id=thread_id,
            run_id=event.run_id,
            step_id=event.step_id,
            reason=event.reason,
            compression_count=event.compression_count,
            included_event_id=included_event_id,
            history=event.history,
        )

    def _archive_subagent_artifact(self, event: AgentEvent) -> AgentEvent:
        if not isinstance(event, ToolCompletedEvent):
            return event
        if event.tool_name != "delegate_to_subagent":
            return event
        session_id = self.session_id
        if session_id is None:
            raise RuntimeError("subagent artifact archive requires an active session")
        archived_result = self._subagent_artifacts.archive_tool_result(
            session_id=session_id,
            parent_run_id=event.run_id,
            tool_result=event.tool_result,
        )
        return ToolCompletedEvent(
            run_id=event.run_id,
            step_index=event.step_index,
            step_id=event.step_id,
            tool_call=event.tool_call,
            tool_name=event.tool_name,
            tool_result=archived_result,
        )

    def _request_meta_for_run(self) -> LLMRequestMeta:
        return {**self._request_meta, "model": self._selected_model}

    def start_child_thread(
        self,
        *,
        task: str,
        instructions: str,
        context_brief: str,
        allowed_tools: list[str],
        child_agent: Agent,
        agent_id: str,
        run_id: str,
        child_question: str,
    ) -> str:
        self.ensure_session()
        thread = self._session_service.create_subagent_thread(
            task=task,
            agent_id=agent_id,
            parent_run_id=self._current_run().run_id,
            instructions=instructions,
            context_brief=context_brief,
            allowed_tools=allowed_tools,
        )
        self._thread_agents[thread.thread_id] = child_agent
        event_id = self._session_service.record_user_message(
            child_question,
            run_id=run_id,
            thread_id=thread.thread_id,
            agent_id=agent_id,
        )
        self._child_context_event_ids[thread.thread_id] = event_id
        self._last_context_event_ids[thread.thread_id] = event_id
        return thread.thread_id

    async def run_child_thread(
        self,
        *,
        task: str,
        instructions: str,
        context_brief: str,
        allowed_tools: list[str],
        child_agent: Agent,
        child_run: AgentRunContext,
        child_question: str,
    ) -> ChildAgentResult:
        thread_id = self.start_child_thread(
            task=task,
            instructions=instructions,
            context_brief=context_brief,
            allowed_tools=allowed_tools,
            child_agent=child_agent,
            agent_id=child_agent.agent_id,
            run_id=child_run.run_id,
            child_question=child_question,
        )
        loop = asyncio.get_running_loop()
        handoff: asyncio.Future[ChildAgentResult] = loop.create_future()
        runtime_task = asyncio.create_task(
            self._run_child_thread_runtime(
                thread_id=thread_id,
                child_agent=child_agent,
                child_run=child_run,
                handoff=handoff,
            )
        )
        self._child_runtimes[thread_id] = ChildThreadRuntime(
            thread_id=thread_id,
            agent_id=child_agent.agent_id,
            run=child_run,
            handoff=handoff,
            task=runtime_task,
        )
        runtime_task.add_done_callback(
            lambda completed_task: self._complete_child_runtime_task(
                thread_id=thread_id,
                task=completed_task,
                handoff=handoff,
            )
        )
        try:
            return await handoff
        except asyncio.CancelledError:
            runtime = self._child_runtimes.get(thread_id)
            task_to_cancel = runtime.task if runtime is not None else runtime_task
            if not task_to_cancel.done():
                task_to_cancel.cancel()
            self.finish_child_thread(
                thread_id=thread_id,
                status="failed",
            )
            try:
                await task_to_cancel
            except asyncio.CancelledError:
                pass
            raise

    async def _run_child_thread_runtime(
        self,
        *,
        thread_id: str,
        child_agent: Agent,
        child_run: AgentRunContext,
        handoff: asyncio.Future[ChildAgentResult],
    ) -> None:
        events: list[AgentEvent] = []
        final_answer = ""
        try:
            async for event in child_agent.execute(child_run):
                events.append(event)
                self.record_child_event(
                    thread_id=thread_id,
                    agent_id=child_agent.agent_id,
                    event=event,
                )
                if isinstance(event, RunCompletedEvent):
                    final_answer = event.final_answer
                    self.finish_child_thread(
                        thread_id=thread_id,
                        status=child_run.status,
                    )
                    if not handoff.done():
                        handoff.set_result(
                            build_child_agent_result(
                                thread_id=thread_id,
                                child_agent=child_agent,
                                child_run=child_run,
                                events=events,
                                final_answer=final_answer,
                            )
                        )
                    return
                if isinstance(event, RunSuspendedEvent):
                    self.finish_child_thread(
                        thread_id=thread_id,
                        status="suspended",
                    )
                    if not handoff.done():
                        handoff.set_exception(
                            ToolExecutionError(
                                "delegated subagent suspended for approval; "
                                "delegate_to_subagent only supports approval-free child tools"
                            )
                        )
                    return
                if isinstance(event, RunFailedEvent):
                    self.finish_child_thread(
                        thread_id=thread_id,
                        status="failed",
                    )
                    if not handoff.done():
                        handoff.set_exception(ToolExecutionError(event.error))
                    return
        except asyncio.CancelledError:
            if child_run.run_id in self._steered_child_run_ids:
                self._steered_child_run_ids.remove(child_run.run_id)
                raise
            error = "delegated subagent run was cancelled before a terminal event"
            child_run.run_state.status = "failed"
            event = _cancelled_child_run_failed_event(
                child_run=child_run,
                events=events,
                error=error,
            )
            events.append(event)
            self.record_child_event(
                thread_id=thread_id,
                agent_id=child_agent.agent_id,
                event=event,
            )
            self.finish_child_thread(
                thread_id=thread_id,
                status="failed",
            )
            if not handoff.done():
                handoff.set_exception(ToolExecutionError(error))
            raise
        if child_run.status == "running":
            if child_run.run_id in self._steered_child_run_ids:
                self._steered_child_run_ids.remove(child_run.run_id)
                return
            error = "delegated subagent run was cancelled before a terminal event"
            child_run.run_state.status = "failed"
            event = _cancelled_child_run_failed_event(
                child_run=child_run,
                events=events,
                error=error,
            )
            events.append(event)
            self.record_child_event(
                thread_id=thread_id,
                agent_id=child_agent.agent_id,
                event=event,
            )
            self.finish_child_thread(
                thread_id=thread_id,
                status="failed",
            )
            if not handoff.done():
                handoff.set_exception(ToolExecutionError(error))
            return
        self.finish_child_thread(
            thread_id=thread_id,
            status=child_run.status,
        )
        if not handoff.done():
            handoff.set_result(
                build_child_agent_result(
                    thread_id=thread_id,
                    child_agent=child_agent,
                    child_run=child_run,
                    events=events,
                    final_answer=final_answer,
                )
            )

    def _complete_child_runtime_task(
        self,
        *,
        thread_id: str,
        task: asyncio.Task[None],
        handoff: asyncio.Future[ChildAgentResult],
    ) -> None:
        if task in self._stale_child_runtime_tasks:
            self._stale_child_runtime_tasks.remove(task)
            return
        runtime = self._child_runtimes.get(thread_id)
        if runtime is not None and runtime.task is not task:
            return
        self._child_runtimes.pop(thread_id, None)
        if task.cancelled():
            if not handoff.done():
                handoff.cancel()
            return
        exception = task.exception()
        if exception is not None and not handoff.done():
            self.finish_child_thread(
                thread_id=thread_id,
                status="failed",
            )
            handoff.set_exception(exception)

    def _record_child_steer_event(
        self,
        *,
        thread_id: str,
        agent_id: str,
        run_id: str,
        question: str,
    ) -> str:
        event_id = self._session_service.record_session_event(
            SessionEvent(
                event_id=uuid_str(),
                thread_id=thread_id,
                agent_id=agent_id,
                run_id=run_id,
                step_id=None,
                step_index=None,
                kind="user_steer",
                payload={"content": question},
            )
        )
        if event_id is None:
            raise RuntimeError("user_steer did not persist a session event")
        return event_id

    def record_child_event(
        self,
        *,
        thread_id: str,
        agent_id: str,
        event: AgentEvent,
    ) -> None:
        if isinstance(event, ContextCompressedEvent):
            included_event_id = self._child_context_event_ids[thread_id]
            session_id = self._session_service.session_id
            if session_id is None:
                raise RuntimeError("AceAI session is not active")
            self._session_service.context_checkpoint_store.record_checkpoint(
                session_id=session_id,
                thread_id=thread_id,
                run_id=event.run_id,
                step_id=event.step_id,
                reason=event.reason,
                compression_count=event.compression_count,
                included_event_id=included_event_id,
                history=event.history,
            )
            self._emit_app_event(
                thread_id=thread_id,
                agent_id=agent_id,
                event=event,
            )
            return
        if isinstance(
            event,
            ContextCompactionFailedEvent
            | ContextCompactionStartedEvent,
        ):
            self._emit_app_event(
                thread_id=thread_id,
                agent_id=agent_id,
                event=event,
            )
            return
        persisted_event_id = self._session_service.record_agent_event(
            event,
            thread_id=thread_id,
            agent_id=agent_id,
        )
        if persisted_event_id is not None and _agent_event_updates_context(event):
            self._child_context_event_ids[thread_id] = persisted_event_id
            self._last_context_event_ids[thread_id] = persisted_event_id
        self._emit_app_event(
            thread_id=thread_id,
            agent_id=agent_id,
            event=event,
        )

    def finish_child_thread(
        self,
        *,
        thread_id: str,
        status: str,
    ) -> None:
        if status not in ("idle", "running", "suspended", "completed", "failed"):
            raise ValueError("Unsupported child thread status")
        self._session_service.update_thread_status(
            thread_id=thread_id,
            status=status,
        )

    def _emit_app_event(
        self,
        *,
        thread_id: str,
        agent_id: str,
        event: AgentEvent,
    ) -> None:
        app_event = AgentAppEvent(
            thread_id=thread_id,
            agent_id=agent_id,
            event=event,
        )
        for queue in tuple(self._app_event_subscribers):
            queue.put_nowait(app_event)

    def _queued_questions_for_active_thread(self) -> list[str]:
        if self._active_thread_id not in self._queued_questions:
            self._queued_questions[self._active_thread_id] = []
        return self._queued_questions[self._active_thread_id]

    def _agent_for_thread(self, thread_id: str) -> Agent:
        if thread_id in self._thread_agents:
            return self._thread_agents[thread_id]
        agent = self._restore_agent_for_thread(thread_id)
        self._thread_agents[thread_id] = agent
        return agent

    def _restore_agent_for_thread(self, thread_id: str) -> Agent:
        session_id = self.session_id
        if session_id is None:
            raise RuntimeError("AceAI session is not active")
        thread = self._session_service.store.get_thread(session_id, thread_id)
        if thread.role != "subagent":
            raise RuntimeError("Only subagent threads can be restored")
        executor = self._agent.executor
        if not isinstance(executor, Executor):
            raise RuntimeError("Subagent thread restore requires an Executor")
        metadata = thread.metadata
        instructions = metadata["instructions"]
        context_brief = metadata["context_brief"]
        allowed_tools = metadata["allowed_tools"]
        if type(instructions) is not str:
            raise TypeError("subagent thread instructions must be str")
        if type(context_brief) is not str:
            raise TypeError("subagent thread context_brief must be str")
        if type(allowed_tools) is not list:
            raise TypeError("subagent thread allowed_tools must be list")
        for tool_name in allowed_tools:
            if type(tool_name) is not str:
                raise TypeError("subagent thread allowed_tools entries must be str")
        available_tools = [
            tool
            for tool_name, tool in executor.tools.items()
            if tool_name != "delegate_to_subagent"
        ]
        return build_restored_delegated_child_agent(
            llm_service=self._agent.llm_service,
            default_model=self._agent.default_model,
            instructions=instructions,
            allowed_tools=allowed_tools,
            available_tools=available_tools,
            available_hosted_tools=executor.hosted_tools,
            agent_id=thread.agent_id,
        )

    def _history_for_thread(self, thread_id: str) -> list[LLMMessage]:
        if thread_id == MAIN_THREAD_ID:
            return self._llm_history
        session_id = self.session_id
        if session_id is None:
            raise RuntimeError("AceAI session is not active")
        snapshot = self._session_service.snapshot_thread(session_id, thread_id)
        latest_context_event_id = _last_context_source_event_id(snapshot.event_log)
        if latest_context_event_id is not None:
            self._last_context_event_ids[thread_id] = latest_context_event_id
        return snapshot.history


def effective_reasoning_level(
    provider_name: str,
    model: str,
    level: ReasoningLevel,
) -> ReasoningLevel:
    if level == "auto":
        return "auto"
    if not supports_reasoning_effort(provider_name, model):
        return "auto"
    return level


def model_options_text_for(provider_name: str) -> str:
    names = ", ".join(option[1] for option in model_options(provider_name))
    return f"Available models: {names}"


def is_model_supported(provider_name: str, model: str) -> bool:
    return model in supported_models(provider_name)


def resolve_provider_api_key(provider_name: str) -> str:
    """Resolve an API key for the provider from env var, then default auth file.

    Returns "" if neither source has a key.
    """

    env_name = api_key_env(provider_name)
    api_key = os.environ.get(env_name, "")
    if api_key != "":
        return api_key
    return default_api_key_for_provider(provider_name)


def normalize_user_config(
    config: AgentAppConfig,
    *,
    persist: bool = False,
) -> AgentAppConfig:
    """Validate the config and optionally persist it to the project file."""

    next_config = replace_config(config)
    if persist:
        save_config(next_config)
    return next_config


def context_window_for_model(model: str | None) -> int | None:
    if model is None or model == "":
        return None
    return context_window_for_model_any_provider(model)


def supports_reasoning_for_model(model: str | None) -> bool:
    if model is None or model == "":
        return False
    return supports_reasoning_effort_any_provider(model)


def _agent_event_updates_context(event: AgentEvent) -> bool:
    return isinstance(event, LLMCompletedEvent | ToolCompletedEvent | ToolFailedEvent)


def _cancelled_child_run_failed_event(
    *,
    child_run: AgentRunContext,
    events: list[AgentEvent],
    error: str,
) -> RunFailedEvent:
    if events:
        step_index = events[-1].step_index
        step_id = events[-1].step_id
        failed_step = AgentStep(
            step_id=step_id,
            llm_response=LLMResponse(text="", status="failed"),
        )
    elif child_run.steps:
        step_index = len(child_run.steps)
        step_id = child_run.steps[-1].step_id
        failed_step = AgentStep(
            step_id=step_id,
            llm_response=LLMResponse(text="", status="failed"),
        )
    else:
        step_index = len(child_run.steps)
        failed_step = AgentStep(
            llm_response=LLMResponse(text="", status="failed"),
        )
        step_id = failed_step.step_id
    return RunFailedEvent(
        run_id=child_run.run_id,
        step_index=step_index,
        step_id=failed_step.step_id,
        step=failed_step,
        error=error,
    )


def _last_context_source_event_id(event_log: EventLog) -> str | None:
    for event in reversed(event_log.events):
        if event.kind in (
            "assistant_message",
            "assistant_tool_call",
            "tool_result",
            "user_message",
        ):
            return event.event_id
    return None


async def check_for_updates() -> UpdateCheckResult | None:
    current_version = aceai.__version__
    latest_version = await _fetch_latest_package_version()
    if latest_version is None:
        return None
    return UpdateCheckResult(
        current_version=current_version,
        latest_version=latest_version,
    )


async def _fetch_latest_package_version() -> str | None:
    return await asyncio.to_thread(_fetch_latest_package_version_sync)


def _fetch_latest_package_version_sync() -> str | None:
    try:
        with urlopen(PYPI_PROJECT_JSON_URL, timeout=2.0) as response:
            payload = json.loads(response.read().decode())
    except (URLError, TimeoutError):
        return None
    if type(payload) is not dict:
        raise TypeError("PyPI project payload must be a mapping")
    info = payload["info"]
    if type(info) is not dict:
        raise TypeError("PyPI project info must be a mapping")
    version = info["version"]
    if type(version) is not str:
        raise TypeError("PyPI project version must be str")
    return version


def _version_parts(version: str) -> tuple[int, int, int]:
    parts = version.split(".")
    if len(parts) != 3:
        raise ValueError("AceAI version must use x.y.z format")
    return (int(parts[0]), int(parts[1]), int(parts[2]))


def _approved_tool_names_from_event_log(event_log: EventLog) -> set[str]:
    approved_tool_names: set[str] = set()
    for event in event_log.events:
        if event.kind != "tool_approval_resolved":
            continue
        if event.payload["content"] != "approved":
            continue
        tool_name = event.payload["tool_name"]
        if type(tool_name) is not str:
            raise TypeError("tool approval payload tool_name must be str")
        approved_tool_names.add(tool_name)
    return approved_tool_names



from .agent import Agent as Agent
from .executor import DummyExecutor as DummyExecutor
from .executor import Executor as Executor
from .executor import IExecutor as IExecutor
from .executor import LoggingExecutor as LoggingExecutor
from .executor import ToolExecutionError as ToolExecutionError
from .events import ContextCompressedEvent as ContextCompressedEvent
from .models import ToolApprovalDecision as ToolApprovalDecision
from .models import ToolApprovalRequest as ToolApprovalRequest
from .models import ToolExecutionOutput as ToolExecutionOutput
from .run_loop import AgentRunContext as AgentRunContext
from .run_loop import execute_agent_run as execute_agent_run
from .run_loop import resume_agent_approval as resume_agent_approval
from .run_state import AgentRunState as AgentRunState
from .run_state import PendingToolApproval as PendingToolApproval
from .run_state import ToolInvocation as ToolInvocation
from .run_state import ToolRunState as ToolRunState



import json

import pytest

from aceai.agent.app import AceAgentApp
from aceai.agent.memory.context_checkpoint_store import (
    ContextCheckpoint,
    ContextCheckpointStore,
    context_units_payload_from_messages,
    llm_message_from_payload,
    llm_message_to_payload,
)
from aceai.agent.memory.context_history import build_context_history
from aceai.agent.session import EventLog, SessionEvent, SessionStore
from aceai.core.agent import Agent
from aceai.llm import LLMResponse
from aceai.llm.models import (
    LLMMessage,
    LLMToolCall,
    LLMToolCallMessage,
    LLMToolUseMessage,
)

from tests.test_agent_behavior import (
    CompressingLLMService,
    StubExecutor,
    make_stream,
)


def test_context_checkpoint_store_round_trips_latest_checkpoint(tmp_path) -> None:
    store = ContextCheckpointStore(tmp_path / "checkpoints")
    call = LLMToolCall(name="lookup", arguments="{}", call_id="call-1")
    history = [
        LLMMessage.build(role="user", content="hello"),
        LLMToolCallMessage.from_content(content=[], tool_calls=[call]),
        LLMToolUseMessage.from_content(
            content='{"ok":true}',
            name="lookup",
            call_id="call-1",
        ),
    ]

    first = store.record_checkpoint(
        session_id="session-1",
        run_id="run-1",
        step_id="step-1",
        reason="threshold",
        compression_count=1,
        included_event_id="event-1",
        history=[LLMMessage.build(role="user", content="older")],
    )
    second = store.record_checkpoint(
        session_id="session-1",
        run_id="run-2",
        step_id="step-2",
        reason="context_window_retry",
        compression_count=2,
        included_event_id="event-2",
        history=history,
    )

    latest = store.latest_checkpoint("session-1")

    assert latest is not None
    assert latest.checkpoint_id == second.checkpoint_id
    assert latest.checkpoint_id != first.checkpoint_id
    assert latest.reason == "context_window_retry"
    assert latest.included_event_id == "event-2"
    assert latest.history[0] == history[0]
    assert isinstance(latest.history[1], LLMToolCallMessage)
    assert latest.history[1].tool_calls == [call]
    assert isinstance(latest.history[2], LLMToolUseMessage)
    assert latest.history[2].call_id == "call-1"
    assert latest.thread_id == "main"
    assert (tmp_path / "checkpoints" / "session-1.main.checkpoints.jsonl").exists()


def test_context_checkpoint_store_keeps_thread_histories_separate(tmp_path) -> None:
    store = ContextCheckpointStore(tmp_path / "checkpoints")

    store.record_checkpoint(
        session_id="session-1",
        thread_id="main",
        run_id="run-1",
        step_id="step-1",
        reason="threshold",
        compression_count=1,
        included_event_id="event-1",
        history=[LLMMessage.build(role="user", content="main history")],
    )
    store.record_checkpoint(
        session_id="session-1",
        thread_id="child-1",
        run_id="run-2",
        step_id="step-2",
        reason="threshold",
        compression_count=1,
        included_event_id="event-2",
        history=[LLMMessage.build(role="user", content="child history")],
    )

    main_checkpoint = store.latest_checkpoint("session-1", thread_id="main")
    child_checkpoint = store.latest_checkpoint("session-1", thread_id="child-1")

    assert main_checkpoint is not None
    assert child_checkpoint is not None
    assert main_checkpoint.history[0].content[0]["data"] == "main history"
    assert child_checkpoint.history[0].content[0]["data"] == "child history"


def test_llm_message_checkpoint_payload_rejects_unknown_message_type() -> None:
    payload = llm_message_to_payload(LLMMessage.build(role="user", content="hello"))
    payload["message_type"] = "unknown"

    with pytest.raises(ValueError, match="Unsupported context checkpoint message_type"):
        llm_message_from_payload(payload)


def test_context_history_uses_checkpoint_without_changing_transcript_replay() -> None:
    event_log = EventLog(
        [
            _session_event("event-1", "user_message", {"content": "old question"}),
            _session_event(
                "event-2",
                "assistant_message",
                {"content": "old answer"},
            ),
            _session_event("event-3", "user_message", {"content": "new question"}),
            _session_event(
                "event-4",
                "assistant_message",
                {"content": "new answer"},
            ),
        ]
    )
    checkpoint_history = [
        LLMMessage.build(
            role="system",
            content=(
                '<aceai_context_summary scope="prior_runs">\n'
                "summary of old context\n"
                "</aceai_context_summary>"
            ),
        ),
        LLMMessage.build(role="user", content="new question"),
    ]
    checkpoint = ContextCheckpoint(
        checkpoint_id="checkpoint-1",
        session_id="session-1",
        thread_id="main",
        run_id="run-1",
        step_id="step-1",
        reason="threshold",
        compression_count=1,
        included_event_id="event-3",
        message_count=2,
        estimated_tokens=10,
        history=checkpoint_history,
        units=context_units_payload_from_messages(checkpoint_history),
    )

    transcript_history = event_log.replay_llm_history()
    context_history = build_context_history(
        event_log=event_log,
        checkpoint=checkpoint,
    )

    assert [message.content[0]["data"] for message in transcript_history] == [
        "old question",
        "old answer",
        "new question",
        "new answer",
    ]
    assert [message.content[0]["data"] for message in context_history] == [
        (
            '<aceai_context_summary scope="prior_runs">\n'
            "summary of old context\n"
            "</aceai_context_summary>"
        ),
        "new question",
        "new answer",
    ]


def test_context_checkpoint_store_rejects_malformed_payload(tmp_path) -> None:
    checkpoint_dir = tmp_path / "checkpoints"
    checkpoint_dir.mkdir()
    path = checkpoint_dir / "session-1.checkpoints.jsonl"
    path.write_text(
        json.dumps(
            {
                "version": 2,
                "checkpoint_id": "checkpoint-1",
                "session_id": "session-1",
                "run_id": "run-1",
                "step_id": "step-1",
                "reason": "threshold",
                "compression_count": 1,
                "included_event_id": "event-1",
                "message_count": 1,
                "estimated_tokens": 1,
                "units": {"bad": "shape"},
            }
        ),
        encoding="utf-8",
    )
    store = ContextCheckpointStore(checkpoint_dir)

    with pytest.raises(TypeError, match="context checkpoint units must be list"):
        store.latest_checkpoint("session-1")


@pytest.mark.anyio
async def test_agent_app_persists_checkpoint_and_restores_context_history(
    tmp_path,
) -> None:
    session_store = SessionStore(tmp_path / "sessions")
    llm_service = CompressingLLMService(
        make_stream(response=LLMResponse(text="done"), deltas=["done"])
    )
    agent = Agent(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,
        executor=StubExecutor(),
        max_steps=1,
        compress_threshold=1,
    )
    initial_history = [
        LLMMessage.build(role="user", content=f"history message {index}")
        for index in range(10)
    ]
    app = AceAgentApp(
        agent,
        provider_name="openai",
        selected_model="gpt-4o",
        initial_history=initial_history,
        session_store=session_store,
    )

    [event async for event in app.start_turn("new question")]

    session_id = app.session_id
    assert session_id is not None
    checkpoint = app.session_service.context_checkpoint_store.latest_checkpoint(
        session_id,
        thread_id="main",
    )
    assert checkpoint is not None
    assert checkpoint.reason == "threshold"
    assert (
        checkpoint.included_event_id
        == session_store.load_event_log(session_id).events[0].event_id
    )

    restored_app = AceAgentApp(
        Agent(
            prompt="Prompt",
            default_model="gpt-4o",
            llm_service=CompressingLLMService(
                make_stream(response=LLMResponse(text="later"), deltas=["later"])
            ),
            executor=StubExecutor(),
            max_steps=1,
        ),
        provider_name="openai",
        selected_model="gpt-4o",
        session_store=session_store,
        session_id=session_id,
    )
    restored_text = "\n".join(
        part["data"]
        for message in restored_app.llm_history
        for part in message.content
        if part["type"] == "text"
    )

    assert "Earlier discussion summary." in restored_text
    assert "done" in restored_text
    assert "history message 0" not in restored_text


def _session_event(
    event_id: str,
    kind: str,
    payload: dict,
) -> SessionEvent:
    return SessionEvent(
        event_id=event_id,
        session_id="session-1",
        run_id="run-1",
        kind=kind,
        payload=payload,
    )



"""Live runner that bridges the AceAI app facade into the Textual app."""

import asyncio
import os
import sys
from pathlib import Path
from typing import AsyncGenerator, Callable, cast

from msgspec import Struct
from opentelemetry.context import Context
from rapidfuzz import fuzz
from textual.timer import Timer
from textual.widgets import Input
from textual.worker import Worker

from aceai.agent.app import (
    AceAgentApp,
    UpdateCheckResult,
    effective_reasoning_level,
    is_model_supported,
    model_options_text_for,
    normalize_user_config,
    resolve_provider_api_key,
)
from aceai.agent.citations import (
    FileCitationOrigin,
    IdeaCitationOrigin,
    TurnCitation,
)
from aceai.agent.memory.ideas import Idea, IdeaStore
from aceai.agent.project import ProjectMetadata
from aceai.agent.features import default_agent_tools
from aceai.agent.session import MAIN_THREAD_ID, SessionRecorder, SessionState
from aceai.agent.ace_agent import ACE_AGENT_BUILTIN_SKILL_PATHS
from aceai.agent.config import (
    AgentAppConfig,
    ReasoningLevel,
    effective_config_path,
    load_config_audit,
)
from aceai.core import Agent
from aceai.core.events import AgentEvent, RunSuspendedEvent
from aceai.core.skills import SkillLoader, SkillLoadingError, SkillRegistry
from aceai.llm.models import LLMMessage
from aceai.llm.openai import OpenAIModel

from .app import AceAITUI
from .events import TUIEvent, TUIIdeaItem
from .metadata import MetadataSection
from .session_replay import event_log_to_tui_events
from .setup import (
    ConfigSelection,
    ConfigScreen,
    IdeaPickerScreen,
    ProviderSetupScreen,
    SkillConfigItem,
    ToolPermissionItem,
    _skill_source,
)
from .widgets import ApprovalWidget
from .widgets import CitationPreviewWidget
from .widgets import CommandCompletionItem
from .widgets import CommandInput
from .widgets import QueuedTurnsWidget
from .widgets import ReferenceCompletionItem
from .widgets import ReferenceCompletionWidget
from .widgets import StatusBarWidget
from .widgets import TopBarWidget

AgentFactory = Callable[[AgentAppConfig], Agent]
CommandHandler = Callable[[str], None]
COMMAND_NAMES_ATTR = "_aceai_tui_command_names"
UPDATE_INSTRUCTIONS = "Run /update to upgrade AceAI and restart."
UPDATE_COMMAND: tuple[str, ...] = ("uv", "tool", "upgrade", "aceai")
REFERENCE_IGNORED_DIRS = {
    ".cache",
    ".git",
    ".mypy_cache",
    ".pytest_cache",
    ".ruff_cache",
    ".tox",
    ".venv",
    "__pycache__",
    "build",
    "dist",
    "node_modules",
    "wheels",
}
COMMAND_DESCRIPTIONS: dict[str, str] = {
    "clear": "Clear the visible transcript",
    "config": "Open provider, model, skill, and tool settings",
    "debug": "Toggle the debug detail view",
    "idea": "Show ideas, save an idea, or delete one",
    "quit": "Exit AceAI",
    "sessions": "Open the session picker",
    "stats": "Open runtime and usage stats",
    "steer": "Interrupt or redirect the current run",
    "subagents": "Show delegated subagent details",
    "trajectory": "Open the event trajectory view",
    "update": "Upgrade AceAI and restart",
}


def tui_command(*names: str):
    if not names:
        raise ValueError("TUI command must declare at least one name")
    for name in names:
        if name == "":
            raise ValueError("TUI command name cannot be empty")
        if name.startswith("/"):
            raise ValueError("TUI command name must not include slash")

    def decorate(handler):
        setattr(handler, COMMAND_NAMES_ATTR, names)
        return handler

    return decorate


def _skill_config_items(registry: SkillRegistry) -> tuple[SkillConfigItem, ...]:
    return tuple(
        SkillConfigItem(
            name=skill.name,
            description=skill.description,
            location=str(skill.skill_file),
            builtin=_is_builtin_skill_location(skill.skill_file),
            source=_skill_source(skill.skill_file),
        )
        for skill in registry.get_skills()
    )


def _is_builtin_skill_location(skill_file: Path) -> bool:
    return _skill_source(skill_file) == "aceai builtin"


def _parse_command(text: str) -> tuple[str, str] | None:
    if not text.startswith("/"):
        return None
    body = text.removeprefix("/")
    if body == "":
        return None
    name, separator, arg = body.partition(" ")
    if separator == "":
        return name, ""
    return name, arg


class AceAIConfiguredTUI(AceAITUI):
    """TUI that asks for provider settings before creating the agent."""

    def __init__(
        self,
        agent_factory: AgentFactory,
        *,
        initial_config: AgentAppConfig | None,
        initial_question: str,
        default_model: OpenAIModel,
        initial_events: list[TUIEvent] | None = None,
        initial_history: list[LLMMessage] | None = None,
        session_recorder: SessionRecorder | None = None,
        session_id: str | None = None,
        project: ProjectMetadata | None = None,
        idea_store: IdeaStore | None = None,
        trace_ctx: Context | None = None,
        reasoning_level: ReasoningLevel = "auto",
        inline_viewport_height: int | None = None,
    ) -> None:
        self._provider_name = (
            initial_config.provider if initial_config is not None else "openai"
        )
        self._current_config = initial_config
        initial_model = (
            initial_config.model if initial_config is not None else default_model
        )
        requested_level: ReasoningLevel = (
            initial_config.reasoning_level
            if initial_config is not None
            else reasoning_level
        )
        self._reasoning_level: ReasoningLevel = effective_reasoning_level(
            self._provider_name,
            initial_model,
            requested_level,
        )
        super().__init__(
            events=initial_events or [],
            model=initial_model,
            reasoning_level=self._reasoning_level,
            session_recorder=session_recorder,
            session_id=session_id,
            project=project,
            idea_store=idea_store,
            record_events=False,
            inline_viewport_height=inline_viewport_height,
        )
        self._agent_factory = agent_factory
        self._initial_config = initial_config
        self._initial_question = initial_question
        self._default_model: OpenAIModel = default_model
        self._trace_ctx = trace_ctx
        self._selected_model: OpenAIModel = initial_model
        self._llm_history = list(initial_history or [])
        self._agent: Agent | None = None
        self._agent_app: AceAgentApp | None = None
        self._active_worker: Worker[None] | None = None
        self._active_run = None
        self._cancel_armed = False
        self._cancel_arm_timer: Timer | None = None
        self._reference_completion_selected_index = 0
        self._reference_file_items_cache_root: Path | None = None
        self._reference_file_items_cache: tuple[ReferenceCompletionItem, ...] = ()

    def _start_update_check(self) -> None:
        if self._agent_app is None:
            return
        self.run_worker(
            self._check_for_updates(),
            name="aceai-update-check",
            description="Check whether a newer AceAI release is available",
            exit_on_error=False,
        )

    async def _check_for_updates(self) -> None:
        result = await self._agent_app.check_for_updates()
        if result is None:
            return
        self.append_event(TUIEvent.session_notice(_update_available_notice(result)))

    async def _stream_agent_turn(
        self,
        question: str,
        citations: tuple[TurnCitation, ...] = (),
    ) -> None:
        if self._agent_app is None:
            raise RuntimeError("AceAI app is not configured")
        await self._consume_agent_stream(
            self._agent_app.start_turn(question, citations=citations)
        )
        self._start_next_queued_run()

    async def _stream_approval_decision(
        self, *, approved: bool, reason: str = ""
    ) -> None:
        if approved:
            stream = self._agent_app.approve_tool()
        else:
            stream = self._agent_app.reject_tool(reason)
        await self._consume_agent_stream(stream)
        self._start_next_queued_run()

    async def _consume_agent_stream(
        self,
        stream: AsyncGenerator[AgentEvent, None],
    ) -> None:
        provider_name = (
            self._agent_app.provider_name if self._agent_app is not None else None
        )
        try:
            async for event in stream:
                self.append_agent_event(event, provider_name=provider_name)
                if isinstance(event, RunSuspendedEvent):
                    self.show_pending_approval()
        finally:
            await stream.aclose()
            self._sync_app_state()

    def _sync_app_state(self) -> None:
        if self._agent_app is None:
            return
        self._session_recorder = self._agent_app.session_recorder
        self._session_id = self._agent_app.session_id
        self._llm_history = self._agent_app.llm_history
        self._active_run = self._agent_app.active_run
        self._active_thread_id = self._agent_app.active_thread_id
        self._refresh_queued_turns()
        if self._session_id is not None:
            self.title = self._window_title()
            if self.is_mounted:
                self.query_one(TopBarWidget).set_title(self.title)

    def on_unmount(self) -> None:
        agent_app = self._agent_app
        if agent_app is not None:
            agent_app.session_service.finalize()
        elif self._session_recorder is not None:
            self._session_recorder.finalize()
        super().on_unmount()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        if not isinstance(event.input, CommandInput):
            return
        self._handle_command_input_submitted(event.input, event.value)

    def on_command_input_submitted(self, event: CommandInput.Submitted) -> None:
        self._handle_command_input_submitted(event.input, event.value)

    def on_text_area_changed(self, event) -> None:
        super().on_text_area_changed(event)
        if not isinstance(event.text_area, CommandInput):
            return
        self._reference_completion_selected_index = 0
        self._refresh_reference_completions(event.text_area.value)

    def on_command_input_reference_completion_requested(
        self,
        event: CommandInput.ReferenceCompletionRequested,
    ) -> None:
        matches = self._matching_reference_items(event.input.value)
        if not matches:
            return
        index = min(self._reference_completion_selected_index, len(matches) - 1)
        event.input.value = _replace_active_reference(
            event.input.value,
            matches[index].value,
        )
        self._reference_completion_selected_index = 0
        self._refresh_reference_completions(event.input.value)
        event.stop()

    def on_command_input_reference_completion_navigation_requested(
        self,
        event: CommandInput.ReferenceCompletionNavigationRequested,
    ) -> None:
        command_input = self.query_one(CommandInput)
        matches = self._matching_reference_items(command_input.value)
        if not matches:
            return
        self._reference_completion_selected_index = (
            self._reference_completion_selected_index + event.direction
        ) % len(matches)
        self._refresh_reference_completions(command_input.value)
        event.stop()

    def _handle_command_input_submitted(
        self,
        command_input: CommandInput,
        value: str,
    ) -> None:
        question = value
        if question == "":
            return
        if self._dispatch_approval_input(question):
            self.exit_command_input(command_input)
            return
        if self._dispatch_command(question):
            self.exit_command_input(command_input)
            return
        try:
            inline_citations = self._inline_citations(question)
        except (OSError, UnicodeDecodeError, ValueError) as exc:
            self.notify_session(str(exc))
            return
        citations = (*self._consume_pending_citations(), *inline_citations)
        self.start_run(question, citations=citations)
        self.exit_command_input(command_input)

    def _dispatch_approval_input(self, text: str) -> bool:
        agent_app = self._agent_app
        if agent_app is None or not agent_app.is_running_suspended:
            return False
        if text in ("a", "approve"):
            self.approve_pending_tool()
            return True
        if text in ("r", "reject"):
            self.reject_pending_tool("rejected by caller")
            return True
        return False

    def cancel_active_run(self) -> bool:
        if self._active_worker is None or not self._active_worker.is_running:
            self._clear_cancel_arm()
            return False
        if not self._cancel_armed:
            self._arm_cancel()
            return True
        self._clear_cancel_arm()
        self._active_worker.cancel()
        self._active_worker = None
        if self._agent_app is not None:
            self._agent_app.cancel_active_turn()
        self.clear_approval_request()
        self._sync_app_state()
        self.append_event(TUIEvent.run_cancelled("Cancelled current response."))
        return True

    def _arm_cancel(self) -> None:
        self._cancel_armed = True
        if self._cancel_arm_timer is not None:
            self._cancel_arm_timer.stop()
        self.query_one(StatusBarWidget).show_notice(
            "Esc again stops response",
            timeout=1.4,
            style="bold #bf616a",
        )
        self._cancel_arm_timer = self.set_timer(1.4, self._clear_cancel_arm)

    def _clear_cancel_arm(self) -> None:
        self._cancel_armed = False
        timer = self._cancel_arm_timer
        self._cancel_arm_timer = None
        if timer is not None:
            timer.stop()

    def _dispatch_command(self, text: str) -> bool:
        parsed = _parse_command(text)
        if parsed is None:
            return False
        name, arg = parsed
        command = self._command_handlers().get(name)
        if command is None:
            return False
        command(arg)
        return True

    def _command_handlers(self) -> dict[str, CommandHandler]:
        handlers: dict[str, CommandHandler] = {}
        for cls in reversed(type(self).mro()):
            for value in cls.__dict__.values():
                names = getattr(value, COMMAND_NAMES_ATTR, None)
                if names is None:
                    continue
                bound = value.__get__(self, type(self))
                for name in names:
                    handlers[name] = bound
        return handlers

    def command_names(self) -> tuple[str, ...]:
        return tuple(sorted(self._command_handlers()))

    def command_completion_items(self) -> tuple[CommandCompletionItem, ...]:
        return tuple(
            CommandCompletionItem(
                command=name,
                description=COMMAND_DESCRIPTIONS[name],
            )
            for name in self.command_names()
        )

    def _refresh_reference_completions(self, value: str) -> None:
        widget = self._reference_completion_widget()
        if widget is None:
            return
        matches = self._matching_reference_items(value)
        if not matches:
            widget.hide()
            return
        if self._reference_completion_selected_index >= len(matches):
            self._reference_completion_selected_index = 0
        widget.show_references(
            list(matches),
            selected_index=self._reference_completion_selected_index,
        )

    def _matching_reference_items(
        self,
        value: str,
    ) -> tuple[ReferenceCompletionItem, ...]:
        prefix = _active_reference_prefix(value)
        if prefix is None:
            return ()
        if prefix == "":
            return ()
        search_items = self._idea_reference_items()
        if not prefix.startswith("idea:"):
            search_items = (*self._file_reference_items(), *search_items)
        ranked_items = _rank_reference_items(prefix, search_items)
        return _filter_ranked_reference_items(ranked_items)

    def _idea_reference_items(self) -> tuple[ReferenceCompletionItem, ...]:
        return tuple(
            ReferenceCompletionItem(
                value=f"@idea:{index}",
                description=_reference_idea_description(idea),
            )
            for index, idea in enumerate(self._list_ideas(), start=1)
        )

    def _file_reference_items(self) -> tuple[ReferenceCompletionItem, ...]:
        root = Path(self._project.root_path)
        if not root.is_dir():
            return ()
        if self._reference_file_items_cache_root == root:
            return self._reference_file_items_cache
        items: list[ReferenceCompletionItem] = []
        for path in _iter_reference_file_candidates(root):
            items.append(
                ReferenceCompletionItem(
                    value="@" + path.relative_to(root).as_posix(),
                    description="file",
                )
            )
        self._reference_file_items_cache_root = root
        self._reference_file_items_cache = tuple(items)
        return self._reference_file_items_cache

    def _reference_completion_widget(self) -> ReferenceCompletionWidget | None:
        matches = list(self.query(ReferenceCompletionWidget))
        if not matches:
            return None
        return matches[0]

    @tui_command("quit")
    def _command_quit(self, arg: str) -> None:
        self.exit()

    @tui_command("clear")
    def _command_clear(self, arg: str) -> None:
        self.load_events([])

    @tui_command("sessions")
    def _command_sessions(self, arg: str) -> None:
        self.open_session_selector()

    @tui_command("stats")
    def _command_stats(self, arg: str) -> None:
        self.open_stats_screen()

    @tui_command("config")
    def _command_config(self, arg: str) -> None:
        self.open_config_screen()

    @tui_command("debug")
    def _command_debug(self, arg: str) -> None:
        self.action_toggle_debug_mode()

    @tui_command("subagents")
    def _command_subagents(self, arg: str) -> None:
        if arg != "":
            if arg == "main":
                self.switch_thread(MAIN_THREAD_ID)
                return
            self.switch_thread(arg)
            return
        self.action_show_subagents()

    @tui_command("trajectory")
    def _command_trajectory(self, arg: str) -> None:
        self.open_trajectory_screen()

    @tui_command("update")
    def _command_update(self, arg: str) -> None:
        if self._active_worker is not None and self._active_worker.is_running:
            self._active_worker.cancel()
        self.append_event(
            TUIEvent.session_notice("Updating AceAI with uv tool upgrade aceai...")
        )
        self.run_worker(
            self._run_update_and_restart(),
            name="aceai-self-update",
            description="Upgrade AceAI and restart this process",
            exit_on_error=False,
        )

    @tui_command("steer")
    def _command_steer(self, arg: str) -> None:
        if arg == "":
            self.append_event(TUIEvent.session_notice("Usage: /steer <message>"))
            return
        self.steer_run(arg)

    @tui_command("idea")
    def _command_idea(self, arg: str) -> None:
        if arg == "":
            self._show_ideas()
            return
        try:
            delete_index = _idea_delete_index(arg)
        except ValueError:
            self.notify_session("Usage: /idea delete <number>")
            return
        if delete_index is not None:
            self._delete_idea(delete_index)
            return
        idea = self._capture_idea(arg)
        self.notify_session(
            f"Saved idea from {idea.created_at.strftime('%Y-%m-%d %H:%M')}."
        )

    def enqueue_run(
        self,
        question: str,
        *,
        citations: tuple[TurnCitation, ...] = (),
    ) -> None:
        self._clear_cancel_arm()
        agent_app = self._agent_app
        if agent_app is None:
            self.append_event(
                TUIEvent.session_notice("Configure AceAI before enqueueing a run.")
            )
            return
        if not agent_app.active_thread_accepts_user_turn:
            if agent_app.steer_active_child_thread(question):
                self.append_event(TUIEvent.user_message(question))
                self.append_event(TUIEvent.session_notice("Steered running subagent."))
            return
        if (
            self._active_worker is None or not self._active_worker.is_running
        ) and not agent_app.is_running_suspended:
            self._start_run_now(question, citations=citations)
            return
        agent_app.enqueue_turn(question)
        if citations:
            self._queued_turn_citations()[question].append(citations)
        self._refresh_queued_turns()

    def steer_run(self, question: str) -> None:
        self._clear_cancel_arm()
        agent_app = self._agent_app
        if agent_app is None:
            self.append_event(
                TUIEvent.session_notice("Configure AceAI before steering a run.")
            )
            return
        if not agent_app.active_thread_accepts_user_turn:
            if agent_app.steer_active_child_thread(question):
                self.append_event(TUIEvent.user_message(question))
                self.append_event(TUIEvent.session_notice("Steered running subagent."))
            return
        if agent_app.is_running_suspended:
            self.append_event(
                TUIEvent.session_notice(
                    "Choose Approve or Reject before steering this run."
                )
            )
            return
        if self._active_worker is not None and self._active_worker.is_running:
            self._active_worker.cancel()
        self._start_run_now(question)

    def steer_queued_run(self, index: int) -> None:
        self._clear_cancel_arm()
        agent_app = self._agent_app
        if agent_app is None:
            self.append_event(
                TUIEvent.session_notice("Configure AceAI before steering a run.")
            )
            return
        if agent_app.is_running_suspended:
            self.append_event(
                TUIEvent.session_notice(
                    "Choose Approve or Reject before steering this run."
                )
            )
            return
        try:
            question = agent_app.take_queued_turn(index)
        except IndexError:
            self._refresh_queued_turns()
            return
        citations = self._pop_queued_turn_citations(question)
        self._refresh_queued_turns()
        if self._active_worker is not None and self._active_worker.is_running:
            self._active_worker.cancel()
        self._start_run_now(question, citations=citations)

    def _start_run_now(
        self,
        question: str,
        *,
        citations: tuple[TurnCitation, ...] = (),
    ) -> None:
        self._clear_cancel_arm()
        if self._agent_app is None:
            raise RuntimeError("AceAI app is not configured")
        self._agent_app.ensure_session()
        self._sync_app_state()
        self._persist_session_state()
        if not citations:
            citations = self._consume_pending_citations()
        self.append_event(TUIEvent.user_message(question, citations=citations))
        self._active_worker = self.run_worker(
            self._stream_agent_turn(question, citations),
            name="aceai-agent",
            description="Run AceAI agent and stream events into the TUI",
            exit_on_error=True,
        )
        self._refresh_citation_preview()

    def _start_next_queued_run(self) -> None:
        agent_app = self._agent_app
        if agent_app is None:
            return
        if agent_app.is_running_suspended:
            return
        question = agent_app.pop_queued_turn()
        if question is None:
            return
        citations = self._pop_queued_turn_citations(question)
        self._refresh_queued_turns()
        self.call_after_refresh(
            lambda: self._start_run_now(question, citations=citations)
        )

    def _refresh_queued_turns(self) -> None:
        if not self.is_mounted:
            return
        agent_app = self._agent_app
        questions = () if agent_app is None else agent_app.queued_questions
        widgets = list(self.query(QueuedTurnsWidget))
        if not widgets:
            return
        widgets[0].set_questions(questions)

    def on_queued_turns_widget_selected(
        self,
        event: QueuedTurnsWidget.Selected,
    ) -> None:
        event.stop()
        self.steer_queued_run(event.index)

    def on_queued_turns_widget_cancelled(
        self,
        event: QueuedTurnsWidget.Cancelled,
    ) -> None:
        event.stop()
        self.cancel_queued_run(event.index)

    def cancel_queued_run(self, index: int) -> None:
        self._clear_cancel_arm()
        agent_app = self._agent_app
        if agent_app is None:
            self.append_event(
                TUIEvent.session_notice("Configure AceAI before cancelling a queued run.")
            )
            return
        try:
            question = agent_app.cancel_queued_turn(index)
        except IndexError:
            self._refresh_queued_turns()
            return
        self._pop_queued_turn_citations(question)
        self._refresh_queued_turns()
        self.append_event(
            TUIEvent.session_notice(
                f"Cancelled queued message {index + 1}: {question}"
            )
        )

    def _capture_idea(self, content: str) -> Idea:
        agent_app = self._agent_app
        if agent_app is None:
            return self._idea_store.capture(
                content,
                project=self._project,
                source_session_id=self._session_id,
            )
        return agent_app.capture_idea(content)

    def _list_ideas(self) -> list[Idea]:
        agent_app = self._agent_app
        if agent_app is None:
            return self._idea_store.list_for_display(current_project=self._project)
        return agent_app.list_ideas()

    def _show_ideas(self) -> None:
        self.push_screen(
            IdeaPickerScreen(
                ideas=self._list_ideas(),
                capture_idea=self._capture_idea_and_list,
                save_idea=self._update_idea_and_list,
                delete_idea=self._delete_idea_and_list,
            ),
            self._reference_idea,
        )

    def action_ideas(self) -> None:
        self._show_ideas()

    def _reference_idea(self, idea: Idea | None) -> None:
        if idea is None:
            self.query_one(CommandInput).focus()
            return
        self._pending_turn_citations = [
            TurnCitation(
                content=idea.content,
                origin=IdeaCitationOrigin(kind="idea", idea_id=idea.idea_id),
            )
        ]
        self._refresh_citation_preview()
        self.query_one(CommandInput).focus()

    def _pending_citations(self) -> list[TurnCitation]:
        if not hasattr(self, "_pending_turn_citations"):
            self._pending_turn_citations = []
        return self._pending_turn_citations

    def _consume_pending_citations(self) -> tuple[TurnCitation, ...]:
        citations = tuple(self._pending_citations())
        self._pending_turn_citations = []
        return citations

    def _queued_turn_citations(self) -> dict[str, list[tuple[TurnCitation, ...]]]:
        if not hasattr(self, "_queued_turn_citation_map"):
            self._queued_turn_citation_map = {}
        return self._queued_turn_citation_map

    def _pop_queued_turn_citations(self, question: str) -> tuple[TurnCitation, ...]:
        queued = self._queued_turn_citations()
        values = queued.get(question)
        if not values:
            return ()
        citations = values.pop(0)
        if not values:
            queued.pop(question, None)
        return citations

    def _inline_citations(self, question: str) -> tuple[TurnCitation, ...]:
        citations: list[TurnCitation] = []
        for reference in _inline_references(question):
            if reference.kind == "file":
                citations.append(self._file_citation(reference.value))
            elif reference.kind == "idea":
                citations.append(self._idea_citation(reference.value))
        return tuple(citations)

    def _file_citation(self, path_text: str) -> TurnCitation:
        path = Path(path_text).expanduser()
        if not path.is_absolute():
            path = Path(self._project.root_path) / path
        path = path.resolve()
        if not path.is_file():
            raise ValueError(f"File not found: {path}")
        content = path.read_text(encoding="utf-8")
        return TurnCitation(
            content=content,
            origin=FileCitationOrigin(kind="file", path=str(path)),
        )

    def _idea_citation(self, value: str) -> TurnCitation:
        try:
            index = int(value)
        except ValueError:
            idea = self._idea_by_id(value)
        else:
            idea = self._idea_by_display_index(index)
        return TurnCitation(
            content=idea.content,
            origin=IdeaCitationOrigin(kind="idea", idea_id=idea.idea_id),
        )

    def _idea_by_display_index(self, index: int) -> Idea:
        if index < 1:
            raise ValueError("Idea reference index must be one-based")
        ideas = self._list_ideas()
        if index > len(ideas):
            raise ValueError(f"Idea not found: {index}")
        return ideas[index - 1]

    def _idea_by_id(self, idea_id: str) -> Idea:
        for idea in self._list_ideas():
            if idea.idea_id == idea_id:
                return idea
        raise ValueError(f"Idea not found: {idea_id}")

    def _refresh_citation_preview(self) -> None:
        if not self.is_mounted:
            return
        widgets = list(self.query(CitationPreviewWidget))
        if not widgets:
            return
        widgets[0].set_citations(tuple(self._pending_citations()))

    def _capture_idea_and_list(self, content: str) -> list[Idea]:
        self._capture_idea(content)
        return self._list_ideas()

    def _update_idea_and_list(self, index: int, content: str) -> list[Idea]:
        agent_app = self._agent_app
        if agent_app is None:
            self._idea_store.update_displayed(
                index,
                content,
                current_project=self._project,
            )
            return self._idea_store.list_for_display(current_project=self._project)
        agent_app.update_idea(index, content)
        return agent_app.list_ideas()

    def _delete_idea_and_list(self, index: int) -> list[Idea]:
        agent_app = self._agent_app
        if agent_app is None:
            self._idea_store.delete_displayed(
                index,
                current_project=self._project,
            )
            return self._idea_store.list_for_display(current_project=self._project)
        agent_app.delete_idea(index)
        return agent_app.list_ideas()

    def _delete_idea(self, index: int) -> None:
        try:
            agent_app = self._agent_app
            if agent_app is None:
                idea = self._idea_store.delete_displayed(
                    index,
                    current_project=self._project,
                )
            else:
                idea = agent_app.delete_idea(index)
        except IndexError:
            self.notify_session(f"No idea found at {index}.")
            return
        self.append_event(TUIEvent.idea_list(_idea_items(self._list_ideas())))
        self.notify_session(f"Deleted idea {index}: {idea.content}")

    async def _run_update_and_restart(self) -> None:
        result = await run_update_command()
        if result.return_code != 0:
            self.append_event(
                TUIEvent.session_notice(
                    f"AceAI update failed with exit code {result.return_code}.\n"
                    f"{result.output}"
                )
            )
            return
        self.append_event(TUIEvent.session_notice("AceAI updated. Restarting..."))
        restart_current_process()


    def on_mount(self) -> None:
        super().on_mount()
        self._start_update_check()
        if self._initial_config is not None:
            self.apply_config(self._initial_config)
            return
        self.push_screen(
            ProviderSetupScreen(default_model=self._default_model),
            self._handle_setup_config,
        )

    def _handle_setup_config(self, config: AgentAppConfig | None) -> None:
        if config is None:
            return
        self.apply_config(config)

    def apply_config(self, config: AgentAppConfig) -> None:
        next_config = normalize_user_config(config)
        model_changed = next_config.model != self._selected_model
        self._provider_name = next_config.provider
        self._current_config = next_config
        self._selected_model = next_config.model
        self._reasoning_level = next_config.reasoning_level
        self._agent = None
        self._agent_app = None
        self._active_run = None
        self._persist_session_state()
        if model_changed:
            self.reset_status_cache_rate()
        self.set_status_model(
            self._selected_model,
            reasoning_level=self._reasoning_level,
        )
        if self._initial_question != "":
            self.start_run(self._initial_question)

    def start_run(
        self,
        question: str,
        *,
        citations: tuple[TurnCitation, ...] = (),
    ) -> None:
        if self._current_config is None:
            self.query_one(CommandInput).value = question
            return
        if self._active_worker is not None and self._active_worker.is_running:
            self.enqueue_run(question, citations=citations)
            return
        self._ensure_agent_app()
        if self._agent_app.is_running_suspended:
            self.append_event(
                TUIEvent.session_notice(
                    "Choose Approve or Reject before starting another run."
                )
            )
            return
        self._start_run_now(question, citations=citations)

    def _ensure_agent_app(self) -> None:
        if self._agent_app is not None:
            return
        if self._current_config is None:
            raise RuntimeError("AceAI app is not configured")
        self._agent = self._agent_factory(self._current_config)
        self._agent_app = AceAgentApp(
            self._agent,
            provider_name=self._provider_name,
            selected_model=self._selected_model,
            reasoning_level=self._reasoning_level,
            initial_history=self._llm_history,
            session_store=self._session_store(),
            session_recorder=self._session_recorder,
            session_id=self._session_id,
            project=self._project,
            idea_store=self._idea_store,
            trace_ctx=self._trace_ctx,
        )
        new_level = self._agent_app.reasoning_level
        if new_level != self._reasoning_level:
            self._reasoning_level = new_level
            self.set_status_model(
                self._selected_model,
                reasoning_level=self._reasoning_level,
            )
        self._sync_app_state()
        self._start_update_check()

    def switch_session(self, session_id: str) -> None:
        if self._agent_app is None:
            super().switch_session(session_id)
            self._reload_llm_history()
            return
        if session_id == self._session_id:
            return
        try:
            snapshot = self._agent_app.switch_session(session_id)
        except KeyError:
            self.notify_session(f"Session not found: {session_id}")
            return
        self._sync_app_state()
        self.load_events(event_log_to_tui_events(snapshot.event_log))
        self.notify_session(f"Resumed session {snapshot.metadata.session_id}")
        self._restore_session_state()

    def switch_thread(self, thread_id: str) -> None:
        if self._current_config is None:
            AceAITUI.switch_thread(self, thread_id)
            return
        self._ensure_agent_app()
        agent_app = self._agent_app
        if agent_app is None:
            self.append_event(
                TUIEvent.session_notice("Configure AceAI before switching threads.")
            )
            return
        try:
            snapshot = agent_app.switch_thread(thread_id)
        except (KeyError, RuntimeError) as exc:
            self.append_event(TUIEvent.session_notice(str(exc)))
            return
        self._sync_app_state()
        self.load_events(event_log_to_tui_events(snapshot.event_log))
        self.notify_session(f"Switched thread {thread_id}")

    def approve_pending_tool(self) -> None:
        request = self._pending_approval_request()
        if request is None:
            self.append_event(TUIEvent.session_notice("No pending tool approval."))
            return
        self.clear_approval_request()
        self._active_worker = self.run_worker(
            self._stream_approval_decision(approved=True),
            name="aceai-approval",
            description="Resume AceAI agent after tool approval",
            exit_on_error=True,
        )

    def show_pending_approval(self) -> None:
        request = self._pending_approval_request()
        if request is None:
            self.append_event(TUIEvent.session_notice("No pending tool approval."))
            return
        self.show_approval_request(request)

    def on_approval_widget_selected(self, event: ApprovalWidget.Selected) -> None:
        event.stop()
        if event.approved:
            self.approve_pending_tool()
            return
        self.reject_pending_tool("rejected by caller")

    def reject_pending_tool(self, reason: str) -> None:
        request = self._pending_approval_request()
        if request is None:
            self.append_event(TUIEvent.session_notice("No pending tool approval."))
            return
        if reason == "":
            reason = "rejected by caller"
        self.clear_approval_request()
        self._active_worker = self.run_worker(
            self._stream_approval_decision(approved=False, reason=reason),
            name="aceai-approval",
            description="Resume AceAI agent after tool rejection",
            exit_on_error=True,
        )

    def _pending_approval_request(self):
        if self._agent_app is None:
            return None
        return self._agent_app.pending_approval_request()

    def show_model(self) -> None:
        self.append_event(
            TUIEvent.session_notice(
                f"Current model: {self._selected_model}\n{model_options_text_for(self._provider_name)}"
            )
        )

    def action_config(self) -> None:
        self.open_config_screen()

    def open_config_screen(self) -> None:
        api_keys: dict[str, str] = {}
        if self._current_config is not None:
            api_keys.update(self._current_config.api_keys)
            api_keys[self._current_config.provider] = self._current_config.api_key
        self.push_screen(
            ConfigScreen(
                provider_name=self._provider_name,
                current_model=self._selected_model,
                default_model=self._current_config.default_model
                if self._current_config is not None
                else self._selected_model,
                skills=self._current_config.skills
                if self._current_config is not None
                else "",
                skill_items=self._available_skill_items(),
                skill_selection_mode=self._current_config.skill_selection_mode
                if self._current_config is not None
                else "all",
                enabled_skills=tuple(self._current_config.enabled_skills)
                if self._current_config is not None
                else (),
                tool_permission_items=self._available_tool_permission_items(),
                audit_entries=load_config_audit(
                    limit=50,
                    target=effective_config_path(),
                ),
                api_keys=api_keys,
                compress_threshold=self._current_config.compress_threshold
                if self._current_config is not None
                else "100%",
                reasoning_level=self._reasoning_level,
                disabled_providers=tuple(self._current_config.disabled_providers)
                if self._current_config is not None
                else (),
            ),
            self._handle_config_selection,
        )

    def _handle_config_selection(self, selection: ConfigSelection | None) -> None:
        if selection is None:
            return
        if type(selection) is str:
            self.switch_model(selection)
            return
        same_provider = selection.provider == self._provider_name
        if same_provider and selection.api_key == "" and self._current_config is None:
            self.switch_model(selection.model)
            return
        api_key = self._resolve_selection_api_key(selection, same_provider=same_provider)
        if api_key == "":
            self.append_event(
                TUIEvent.session_notice(
                    f"API key required for provider {selection.provider}."
                )
            )
            return
        api_keys = (
            dict(self._current_config.api_keys)
            if self._current_config is not None
            else {}
        )
        api_keys[selection.provider] = api_key
        self.apply_user_config(
            AgentAppConfig(
                provider=selection.provider,
                api_key=api_key,
                model=selection.model,
                default_model=selection.default_model,
                skills=selection.skills,
                skill_selection_mode=selection.skill_selection_mode,
                enabled_skills=list(selection.enabled_skills),
                api_keys=api_keys,
                tool_permissions=selection.tool_permissions,
                tool_enabled=selection.tool_enabled,
                tool_max_calls=selection.tool_max_calls,
                compress_threshold=selection.compress_threshold,
                reasoning_level=selection.reasoning_level,
                disabled_providers=list(selection.disabled_providers),
            )
        )
        if same_provider:
            self.notify_session(
                f"Updated provider credentials and switched model to {selection.model}"
            )
        else:
            self.notify_session(
                f"Switched provider to {selection.provider} and model to {selection.model}"
            )

    def on_config_screen_persist_requested(
        self,
        event: ConfigScreen.PersistRequested,
    ) -> None:
        event.stop()
        self._handle_config_selection(event.selection)

    def _resolve_selection_api_key(
        self,
        selection: ConfigSelection,
        *,
        same_provider: bool,
    ) -> str:
        if selection.api_key != "":
            return selection.api_key
        if same_provider and self._current_config is not None:
            return self._current_config.api_key
        return resolve_provider_api_key(selection.provider)

    def apply_user_config(self, config: AgentAppConfig) -> None:
        normalized = normalize_user_config(config, persist=True)
        self.apply_config(normalized)

    def switch_model(self, model: str) -> None:
        if self._active_worker is not None and self._active_worker.is_running:
            self.append_event(
                TUIEvent.session_notice(
                    "Model changes apply after the current run finishes."
                )
            )
            return
        if not is_model_supported(self._provider_name, model):
            self.append_event(
                TUIEvent.session_notice(model_options_text_for(self._provider_name))
            )
            return
        model_changed = model != self._selected_model
        requested_level = self._reasoning_level
        if self._current_config is not None:
            requested_level = self._current_config.reasoning_level
        next_level = effective_reasoning_level(
            self._provider_name,
            model,
            requested_level,
        )
        if self._current_config is not None:
            next_config = AgentAppConfig(
                provider=self._current_config.provider,
                api_key=self._current_config.api_key,
                model=cast(OpenAIModel, model),
                default_model=cast(OpenAIModel, model),
                skills=self._current_config.skills,
                skill_selection_mode=self._current_config.skill_selection_mode,
                enabled_skills=self._current_config.enabled_skills,
                api_keys=self._current_config.api_keys,
                tool_permissions=self._current_config.tool_permissions,
                tool_enabled=self._current_config.tool_enabled,
                tool_max_calls=self._current_config.tool_max_calls,
                compress_threshold=self._current_config.compress_threshold,
                reasoning_level=next_level,
                disabled_providers=self._current_config.disabled_providers,
            )
            self._current_config = normalize_user_config(next_config, persist=True)
        self._selected_model = cast(OpenAIModel, model)
        self._reasoning_level = next_level
        if self._agent_app is not None:
            self._agent_app.switch_model(
                self._selected_model,
                reasoning_level=self._reasoning_level,
            )
            self._reasoning_level = self._agent_app.reasoning_level
            self._sync_app_state()
        self._persist_session_state()
        if model_changed:
            self.reset_status_cache_rate()
        self.set_status_model(
            self._selected_model,
            reasoning_level=self._reasoning_level,
        )
        self.notify_session(f"Switched model to {self._selected_model}")

    def _available_skill_items(self) -> tuple[SkillConfigItem, ...]:
        if self._current_config is None:
            if self._agent is None:
                return ()
            return _skill_config_items(self._agent.skill_registry)
        try:
            registry = SkillLoader.load_registry(
                self._current_config.skills,
                extra_skill_paths=ACE_AGENT_BUILTIN_SKILL_PATHS,
            )
        except (SkillLoadingError, OSError) as exc:
            self.notify_session(f"Skill search failed: {exc}")
            return ()
        return _skill_config_items(registry)

    def _available_tool_permission_items(self) -> tuple[ToolPermissionItem, ...]:
        configured_permissions = (
            self._current_config.tool_permissions
            if self._current_config is not None
            else {}
        )
        configured_enabled = (
            self._current_config.tool_enabled
            if self._current_config is not None
            else {}
        )
        configured_max_calls = (
            self._current_config.tool_max_calls
            if self._current_config is not None
            else {}
        )
        items: list[ToolPermissionItem] = []
        for configured_tool in default_agent_tools():
            permission = configured_permissions.get(configured_tool.name)
            if permission is None:
                permission = (
                    "ask" if configured_tool.metadata.require_approval else "always"
                )
            items.append(
                ToolPermissionItem(
                    name=configured_tool.name,
                    description=configured_tool.description,
                    permission=permission,
                    enabled=configured_enabled.get(configured_tool.name, True),
                    max_calls_per_run=configured_max_calls.get(configured_tool.name),
                    tags=tuple(configured_tool.metadata.tags),
                )
            )
        return tuple(items)

    def _reload_llm_history(self) -> None:
        if self._session_recorder is None or self._session_id is None:
            self._llm_history = []
            return
        if self._agent_app is None:
            event_log = self._session_recorder.store.load_thread_event_log(
                self._session_id,
                self._active_thread_id,
            )
            self._llm_history = event_log.replay_llm_history()
            self._active_run = None
            return
        self._agent_app.restore_history_from_active_session()
        self._sync_app_state()

    def _persist_session_state(self) -> None:
        if self._session_recorder is None or self._session_id is None:
            return
        if self._agent_app is None:
            self._session_recorder.store.update_session_state(
                self._session_id,
                SessionState(
                    selected_provider=self._provider_name,
                    selected_model=self._selected_model,
                ),
            )
            return
        self._agent_app.persist_session_state()

    def _restore_session_state(self) -> None:
        if self._session_recorder is None or self._session_id is None:
            return
        state = self._session_recorder.store.get_session_state(self._session_id)
        if state.selected_model == "":
            return
        if state.selected_provider == self._provider_name:
            self.switch_model(state.selected_model)

    def _metadata_sections(self) -> list[MetadataSection]:
        sections = super()._metadata_sections()
        if self._agent_app is None:
            return [
                *sections,
                MetadataSection(
                    title="Agent",
                    lines=[
                        f"provider: {self._provider_name}",
                        f"model: {self._selected_model}",
                        "configured: no",
                    ],
                ),
            ]
        return [
            *sections,
            *_agent_metadata_sections(self._agent_app),
        ]


def _update_available_notice(result: UpdateCheckResult) -> str:
    return (
        f"AceAI {result.latest_version} is available "
        f"(current {result.current_version}).\n"
        f"{UPDATE_INSTRUCTIONS}"
    )


def _idea_items(ideas: list[Idea]) -> list[TUIIdeaItem]:
    return [_idea_item(index, idea) for index, idea in enumerate(ideas, start=1)]


def _idea_item(index: int, idea: Idea) -> TUIIdeaItem:
    lines = idea.content.splitlines()
    title = ""
    body_lines: list[str] = []
    for line in lines:
        if title == "" and line != "":
            title = line
            continue
        if title != "":
            body_lines.append(line)
    return TUIIdeaItem(
        index=index,
        project_name=idea.project_name,
        created_at=idea.created_at.strftime("%Y-%m-%d %H:%M"),
        title=title,
        body="\n".join(body_lines),
    )


def _idea_delete_index(arg: str) -> int | None:
    prefix = "delete "
    if not arg.startswith(prefix):
        return None
    return int(arg.removeprefix(prefix))


class InlineReference(Struct, frozen=True, kw_only=True):
    kind: str
    value: str


def _inline_references(text: str) -> tuple[InlineReference, ...]:
    references: list[InlineReference] = []
    seen: set[tuple[str, str]] = set()
    for token in text.split():
        reference = _inline_reference(token)
        if reference is None:
            continue
        key = (reference.kind, reference.value)
        if key in seen:
            continue
        seen.add(key)
        references.append(reference)
    return tuple(references)


def _inline_reference(token: str) -> InlineReference | None:
    token = token.strip()
    if not token.startswith("@") or len(token) == 1:
        return None
    value = token[1:].strip(".,;:!?)]}\"'")
    if value == "" or "://" in value:
        return None
    if value.startswith("idea:"):
        idea = value.removeprefix("idea:")
        if idea == "":
            return None
        return InlineReference(kind="idea", value=idea)
    return InlineReference(kind="file", value=value)


def _active_reference_prefix(value: str) -> str | None:
    tail = value.rsplit(maxsplit=1)[-1] if value.split() else value
    if tail.startswith("@"):
        return tail[1:]
    return None


def _replace_active_reference(value: str, replacement: str) -> str:
    if not value.split():
        return replacement + " "
    head, separator, tail = value.rpartition(" ")
    if not tail.startswith("@"):
        return value
    if separator == "":
        return replacement + " "
    return f"{head} {replacement} "


def _reference_idea_description(idea: Idea) -> str:
    first_line = idea.content.splitlines()[0] if idea.content.splitlines() else "idea"
    if len(first_line) > 64:
        first_line = first_line[:61] + "..."
    return first_line


def _rank_reference_items(
    prefix: str,
    items: tuple[ReferenceCompletionItem, ...],
) -> tuple[tuple[ReferenceCompletionItem, float, int], ...]:
    ranked_items = []
    query = prefix.casefold()
    for index, item in enumerate(items):
        candidate = item.value.removeprefix("@").casefold()
        score = fuzz.WRatio(query, candidate)
        ranked_items.append((item, score, index))
    return tuple(
        sorted(
            ranked_items,
            key=lambda match: (-match[1], match[2]),
        )
    )


def _filter_ranked_reference_items(
    ranked_items: tuple[tuple[ReferenceCompletionItem, float, int], ...],
    *,
    score_cutoff: float = 50,
    limit: int = 12,
) -> tuple[ReferenceCompletionItem, ...]:
    return tuple(
        item
        for item, score, _index in ranked_items
        if score >= score_cutoff
    )[:limit]


def _iter_reference_file_candidates(root: Path):
    for dirpath, dirnames, filenames in os.walk(root):
        dirnames[:] = sorted(
            dirname for dirname in dirnames if dirname not in REFERENCE_IGNORED_DIRS
        )
        for filename in sorted(filenames):
            if filename.startswith("."):
                continue
            path = Path(dirpath) / filename
            if not path.is_file():
                continue
            yield path


class UpdateCommandResult(Struct, frozen=True, kw_only=True):
    return_code: int
    output: str


async def run_update_command() -> UpdateCommandResult:
    try:
        process = await asyncio.create_subprocess_exec(
            *UPDATE_COMMAND,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
    except OSError as err:
        return UpdateCommandResult(return_code=127, output=str(err))
    output, _ = await process.communicate()
    if process.returncode is None:
        raise RuntimeError("uv update process did not report an exit code")
    return UpdateCommandResult(
        return_code=process.returncode,
        output=output.decode(errors="replace"),
    )


def restart_current_process() -> None:
    if not sys.argv:
        os.execv(sys.executable, [sys.executable])
    executable = sys.argv[0]
    if executable == "":
        os.execv(sys.executable, [sys.executable])
    os.execvp(executable, sys.argv)


def _agent_metadata_sections(agent_app: AceAgentApp) -> list[MetadataSection]:
    info = agent_app.runtime_info()
    skill_lines = agent_app.skill_summary_lines()
    tool_lines = agent_app.tool_summary_lines()
    hosted_lines = agent_app.hosted_tool_summary_lines()
    return [
        MetadataSection(
            title="Agent",
            lines=[
                f"provider: {info.provider_name}",
                f"selected model: {info.selected_model}",
                f"default model: {info.default_model}",
                f"max steps: {info.max_steps}",
            ],
        ),
        MetadataSection(title=f"Skills ({len(skill_lines)})", lines=skill_lines),
        MetadataSection(title=f"Tools ({len(tool_lines)})", lines=tool_lines),
        MetadataSection(
            title=f"Hosted Tools ({len(hosted_lines)})",
            lines=hosted_lines,
        ),
    ]



from .delegation import ChildAgentResult as ChildAgentResult
from .delegation import ChildToolResult as ChildToolResult
from .delegation import build_delegated_child_agent as build_delegated_child_agent
from .delegation import build_delegate_to_subagent_tool as build_delegate_to_subagent_tool
from .delegation import build_restored_delegated_child_agent as build_restored_delegated_child_agent
from .tools import default_agent_tools as default_agent_tools



"""Agent app configuration schema, singleton state, and persistence."""

import fcntl
import inspect
import json
import os
from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterator, Literal, TypedDict

import yaml
from msgspec import field

from aceai.agent.provider_catalog import (
    api_key_env,
    default_model,
    model_options,
    reasoning_effort_options,
    stale_default_models,
    supported_models,
    supported_provider_names,
)
from aceai.agent.provider_auth import default_api_key_for_provider
from aceai.agent.permissions import ToolPermission
from aceai.agent.ace_agent import ACE_AGENT_SKILL_PATH
from aceai.core.context_manager import CompressThreshold, ContextCompressionPolicy
from aceai.llm.interface import Record
from aceai.llm.openai import OpenAIModel

ProviderName = str
ConfigValueType = Literal["string", "mapping", "list"]
SkillSelectionMode = Literal["all", "selected"]
ReasoningLevel = Literal["auto", "low", "medium", "high", "max"]
CONFIG_VERSION = 4
DEFAULT_OPENAI_MODEL: OpenAIModel = default_model("openai")
STALE_DEFAULT_OPENAI_MODELS: tuple[OpenAIModel, ...] = stale_default_models("openai")
OPENAI_MODEL_OPTIONS: tuple[tuple[str, OpenAIModel], ...] = model_options("openai")
SUPPORTED_OPENAI_MODELS: tuple[OpenAIModel, ...] = supported_models("openai")
LEGACY_AGENT_SKILLS_DIR = Path(__file__).parent / "features" / "skills"
_PROCESS_START_HOME = Path.home()


class ConfigAuditState(TypedDict):
    provider: object
    model: object
    default_model: object
    skills: object
    skill_selection_mode: object
    enabled_skills: object
    disabled_providers: object
    api_key_providers: list[str]
    tool_permissions: object
    tool_enabled: object
    tool_max_calls: object
    compress_threshold: object
    reasoning_level: object


class ConfigAuditEntry(Record, kw_only=True):
    timestamp: str
    actor: str
    pid: int
    cwd: str
    target: str
    caller: tuple[str, ...]
    changed_fields: tuple[str, ...]
    before: ConfigAuditState | None
    after: ConfigAuditState


class ConfigField(Record, kw_only=True):
    name: str
    value_type: ConfigValueType
    required: bool
    description: str


class ConfigSchema(Record, kw_only=True):
    version: int
    fields: tuple[ConfigField, ...]


class AgentAppConfig(Record, kw_only=True):
    provider: ProviderName
    api_key: str
    model: OpenAIModel
    default_model: OpenAIModel = DEFAULT_OPENAI_MODEL
    skills: str = ACE_AGENT_SKILL_PATH
    skill_selection_mode: SkillSelectionMode = "all"
    enabled_skills: list[str] = field(default_factory=list[str])
    api_keys: dict[str, str] = field(default_factory=dict[str, str])
    tool_permissions: dict[str, ToolPermission] = field(
        default_factory=dict[str, ToolPermission]
    )
    tool_enabled: dict[str, bool] = field(default_factory=dict[str, bool])
    tool_max_calls: dict[str, int] = field(default_factory=dict[str, int])
    compress_threshold: CompressThreshold = "100%"
    reasoning_level: ReasoningLevel = "auto"
    disabled_providers: list[str] = field(default_factory=list[str])


APP_CONFIG_SCHEMA = ConfigSchema(
    version=CONFIG_VERSION,
    fields=(
        ConfigField(
            name="provider",
            value_type="string",
            required=True,
            description="Active model provider.",
        ),
        ConfigField(
            name="model",
            value_type="string",
            required=True,
            description="Active model for new agent turns.",
        ),
        ConfigField(
            name="default_model",
            value_type="string",
            required=True,
            description="Default model used when building the app agent.",
        ),
        ConfigField(
            name="api_key",
            value_type="string",
            required=True,
            description="API key for the active provider.",
        ),
        ConfigField(
            name="skills",
            value_type="string",
            required=True,
            description="Internal skill loading path or mode.",
        ),
        ConfigField(
            name="skill_selection_mode",
            value_type="string",
            required=True,
            description="Whether all loaded skills or selected skills are enabled.",
        ),
        ConfigField(
            name="enabled_skills",
            value_type="list",
            required=True,
            description="Skill names enabled for the active agent model.",
        ),
        ConfigField(
            name="api_keys",
            value_type="mapping",
            required=True,
            description="Provider-key mapping persisted for known providers.",
        ),
        ConfigField(
            name="disabled_providers",
            value_type="list",
            required=True,
            description="Providers that cannot be selected or used.",
        ),
        ConfigField(
            name="tool_permissions",
            value_type="mapping",
            required=True,
            description="Per-tool permission policy: always or ask.",
        ),
        ConfigField(
            name="tool_enabled",
            value_type="mapping",
            required=True,
            description="Per-tool enabled state. Disabled tools are not loaded.",
        ),
        ConfigField(
            name="tool_max_calls",
            value_type="mapping",
            required=True,
            description="Per-tool maximum call count. Missing tools are unlimited.",
        ),
        ConfigField(
            name="compress_threshold",
            value_type="string",
            required=True,
            description="Context compression threshold as a percentage or token count.",
        ),
        ConfigField(
            name="reasoning_level",
            value_type="string",
            required=True,
            description="Reasoning effort level for models that support it.",
        ),
    ),
)

_ACTIVE_CONFIG: AgentAppConfig | None = None


def config_schema() -> ConfigSchema:
    return APP_CONFIG_SCHEMA


def current_config() -> AgentAppConfig | None:
    return _ACTIVE_CONFIG


def replace_config(config: AgentAppConfig) -> AgentAppConfig:
    global _ACTIVE_CONFIG
    validate_config(config)
    old_config = _ACTIVE_CONFIG
    _ACTIVE_CONFIG = config
    del old_config
    return _ACTIVE_CONFIG


def clear_config() -> None:
    global _ACTIVE_CONFIG
    old_config = _ACTIVE_CONFIG
    _ACTIVE_CONFIG = None
    del old_config


def validate_config(config: AgentAppConfig) -> None:
    if config.provider not in supported_provider_names():
        raise ValueError("AceAI config provider is unsupported")
    for provider in config.disabled_providers:
        if type(provider) is not str:
            raise TypeError("AceAI config disabled_providers values must be str")
        if provider not in supported_provider_names():
            raise ValueError("AceAI config disabled_providers value is unsupported")
    if config.provider in config.disabled_providers:
        raise ValueError("AceAI config provider is disabled")
    if config.model not in supported_models(config.provider):
        raise ValueError("AceAI config model is unsupported")
    if config.default_model not in supported_models(config.provider):
        raise ValueError("AceAI config default_model is unsupported")
    if config.reasoning_level not in ("auto", "low", "medium", "high", "max"):
        raise ValueError("AceAI config reasoning_level is unsupported")
    if config.reasoning_level != "auto":
        options = reasoning_effort_options(config.provider, config.model)
        if config.reasoning_level not in options:
            raise ValueError("AceAI config reasoning_level is unsupported for model")
    if config.api_key == "":
        raise ValueError("AceAI config api_key is required")
    if config.skills == "":
        raise ValueError("AceAI config skills is required")
    if config.skill_selection_mode not in ("all", "selected"):
        raise ValueError("AceAI config skill_selection_mode is unsupported")
    for skill_name in config.enabled_skills:
        if type(skill_name) is not str:
            raise TypeError("AceAI config enabled_skills values must be str")
    if config.provider not in config.api_keys:
        raise ValueError("AceAI config api_keys missing active provider")
    if config.api_keys[config.provider] != config.api_key:
        raise ValueError("AceAI config active api_key does not match api_keys")
    for tool_name, permission in config.tool_permissions.items():
        if type(tool_name) is not str:
            raise TypeError("AceAI config tool_permissions tool names must be str")
        if permission not in ("always", "ask"):
            raise ValueError("AceAI config tool_permissions value is unsupported")
    for tool_name, enabled in config.tool_enabled.items():
        if type(tool_name) is not str:
            raise TypeError("AceAI config tool_enabled tool names must be str")
        if type(enabled) is not bool:
            raise TypeError("AceAI config tool_enabled values must be bool")
    for tool_name, max_calls in config.tool_max_calls.items():
        if type(tool_name) is not str:
            raise TypeError("AceAI config tool_max_calls tool names must be str")
        if type(max_calls) is not int:
            raise TypeError("AceAI config tool_max_calls values must be int")
        if max_calls < 1:
            raise ValueError("AceAI config tool_max_calls values must be positive")
    ContextCompressionPolicy(config.compress_threshold)


def default_config_path() -> Path:
    return Path.home() / ".aceai" / "config.yaml"


def project_config_path(project_dir: Path | None = None) -> Path:
    root = project_dir if project_dir is not None else Path.cwd()
    return root / ".aceai" / "config.yml"


def effective_config_path(project_dir: Path | None = None) -> Path:
    project_path = project_config_path(project_dir)
    if project_path.exists():
        return project_path
    return default_config_path()


def config_audit_path() -> Path:
    return default_config_path().parent / "config.audit.jsonl"


def load_config_audit(
    *,
    limit: int = 100,
    target: Path | None = None,
) -> tuple[ConfigAuditEntry, ...]:
    audit_path = config_audit_path()
    if not audit_path.exists():
        return ()
    target_value = str(target) if target is not None else None
    entries: list[ConfigAuditEntry] = []
    with _config_file_lock(audit_path, exclusive=False):
        lines = audit_path.read_text(encoding="utf-8").splitlines()
    for line in lines:
        if line == "":
            continue
        entry = _config_audit_entry_from_mapping(json.loads(line))
        if target_value is not None and entry.target != target_value:
            continue
        entries.append(entry)
    return tuple(reversed(entries[-limit:]))


def load_config(path: Path | None = None) -> AgentAppConfig | None:
    target = path if path is not None else effective_config_path()
    if not target.exists():
        clear_config()
        return None
    with _config_file_lock(target, exclusive=False):
        data = yaml.safe_load(target.read_text(encoding="utf-8"))
    if not isinstance(data, dict):
        raise TypeError("AceAI config must be a mapping")
    provider = data["provider"]
    if provider not in supported_provider_names():
        raise ValueError("AceAI config provider is unsupported")
    api_keys = _load_api_keys(data)
    disabled_providers = _load_disabled_providers(data)
    configured_provider = provider
    selected_provider = _select_loadable_provider(
        provider,
        api_keys,
        disabled_providers,
    )
    if selected_provider is None:
        clear_config()
        return None
    provider = selected_provider
    api_key = _api_key_for_loadable_provider(provider, api_keys)
    api_keys[provider] = api_key
    model = data["model"] if "model" in data else default_model(provider)
    if configured_provider != provider and model not in supported_models(provider):
        model = default_model(provider)
    configured_default_model = (
        data["default_model"] if "default_model" in data else model
    )
    if (
        configured_provider != provider
        and configured_default_model not in supported_models(provider)
    ):
        configured_default_model = default_model(provider)
    skills = data["skills"] if "skills" in data else ACE_AGENT_SKILL_PATH
    legacy_skill_path = False
    if skills == str(LEGACY_AGENT_SKILLS_DIR):
        skills = ACE_AGENT_SKILL_PATH
        legacy_skill_path = True
    skill_selection_mode = (
        data["skill_selection_mode"] if "skill_selection_mode" in data else "all"
    )
    enabled_skills = _load_enabled_skills(data)
    tool_permissions = _load_tool_permissions(data)
    tool_enabled = _load_tool_enabled(data)
    tool_max_calls = _load_tool_max_calls(data)
    compress_threshold = _load_compress_threshold(data)
    reasoning_level = _load_reasoning_level(data)
    if legacy_skill_path:
        skill_selection_mode = "all"
        enabled_skills = []
    has_config_version = "config_version" in data
    if has_config_version and type(data["config_version"]) is not int:
        raise TypeError("AceAI config config_version must be int")
    if model not in supported_models(provider):
        raise ValueError("AceAI config model is unsupported")
    if configured_default_model not in supported_models(provider):
        raise ValueError("AceAI config default_model is unsupported")
    if type(skills) is not str:
        raise TypeError("AceAI config skills must be str")
    if skill_selection_mode not in ("all", "selected"):
        raise ValueError("AceAI config skill_selection_mode is unsupported")
    if not has_config_version and model in stale_default_models(provider):
        model = default_model(provider)
    if not has_config_version and configured_default_model in stale_default_models(
        provider
    ):
        configured_default_model = default_model(provider)
    return replace_config(
        AgentAppConfig(
            provider=provider,
            api_key=api_key,
            model=model,
            default_model=configured_default_model,
            skills=skills,
            skill_selection_mode=skill_selection_mode,
            enabled_skills=enabled_skills,
            api_keys=api_keys,
            tool_permissions=tool_permissions,
            tool_enabled=tool_enabled,
            tool_max_calls=tool_max_calls,
            compress_threshold=compress_threshold,
            reasoning_level=reasoning_level,
            disabled_providers=disabled_providers,
        )
    )


def save_config(config: AgentAppConfig, path: Path | None = None) -> None:
    validate_config(config)
    replace_config(config)
    target = path if path is not None else project_config_path()
    _reject_test_write_to_repo_config(target)
    target.parent.mkdir(parents=True, exist_ok=True)
    with _config_file_lock(target, exclusive=True):
        before_state = _config_audit_state_from_file(target)
        after_state = _config_audit_state_from_config(config)
        payload = yaml.safe_dump(
            {
                "config_version": CONFIG_VERSION,
                "provider": config.provider,
                "api_keys": {**config.api_keys, config.provider: config.api_key},
                "disabled_providers": config.disabled_providers,
                "model": config.model,
                "default_model": config.default_model,
                "skills": config.skills,
                "skill_selection_mode": config.skill_selection_mode,
                "enabled_skills": config.enabled_skills,
                "tool_permissions": config.tool_permissions,
                "tool_enabled": config.tool_enabled,
                "tool_max_calls": config.tool_max_calls,
                "compress_threshold": config.compress_threshold,
                "reasoning_level": config.reasoning_level,
            },
            sort_keys=False,
        )
        with NamedTemporaryFile(
            "w",
            encoding="utf-8",
            dir=target.parent,
            prefix=f".{target.name}.",
            suffix=".tmp",
            delete=False,
        ) as stream:
            stream.write(payload)
            temporary_path = Path(stream.name)
        temporary_path.chmod(0o600)
        temporary_path.replace(target)
        target.chmod(0o600)
        if before_state != after_state:
            _append_config_audit_entry(
                target,
                before=before_state,
                after=after_state,
            )


@contextmanager
def _config_file_lock(target: Path, *, exclusive: bool) -> Iterator[None]:
    lock_path = _config_lock_path(target)
    lock_path.parent.mkdir(parents=True, exist_ok=True)
    operation = fcntl.LOCK_EX if exclusive else fcntl.LOCK_SH
    with lock_path.open("a+", encoding="utf-8") as stream:
        fcntl.flock(stream.fileno(), operation)
        try:
            yield
        finally:
            fcntl.flock(stream.fileno(), fcntl.LOCK_UN)


def _config_lock_path(target: Path) -> Path:
    return target.with_name(f"{target.name}.lock")


def _repo_project_config_path() -> Path:
    return Path(__file__).resolve().parents[2] / ".aceai" / "config.yml"


def _reject_test_write_to_repo_config(target: Path) -> None:
    if "PYTEST_CURRENT_TEST" not in os.environ:
        return
    if target.resolve() != _repo_project_config_path().resolve():
        return
    raise RuntimeError(
        "Tests must not write the real AceAI project config; use tmp_path instead"
    )


def _config_audit_state_from_file(target: Path) -> ConfigAuditState | None:
    if not target.exists():
        return None
    data = yaml.safe_load(target.read_text(encoding="utf-8"))
    if not isinstance(data, dict):
        raise TypeError("AceAI config must be a mapping")
    return _config_audit_state_from_mapping(data)


def _config_audit_state_from_config(config: AgentAppConfig) -> ConfigAuditState:
    return {
        "provider": config.provider,
        "model": config.model,
        "default_model": config.default_model,
        "skills": config.skills,
        "skill_selection_mode": config.skill_selection_mode,
        "enabled_skills": list(config.enabled_skills),
        "disabled_providers": list(config.disabled_providers),
        "api_key_providers": sorted(
            {provider for provider, api_key in config.api_keys.items() if api_key != ""}
            | {config.provider}
        ),
        "tool_permissions": dict(config.tool_permissions),
        "tool_enabled": dict(config.tool_enabled),
        "tool_max_calls": dict(config.tool_max_calls),
        "compress_threshold": config.compress_threshold,
        "reasoning_level": config.reasoning_level,
    }


def _config_audit_state_from_mapping(data: dict[object, object]) -> ConfigAuditState:
    raw_api_keys = data["api_keys"] if "api_keys" in data else {}
    if not isinstance(raw_api_keys, dict):
        raise TypeError("AceAI config api_keys must be a mapping")
    api_key_providers = sorted(
        provider
        for provider, api_key in raw_api_keys.items()
        if type(provider) is str and type(api_key) is str and api_key != ""
    )
    return {
        "provider": data["provider"] if "provider" in data else None,
        "model": data["model"] if "model" in data else None,
        "default_model": data["default_model"] if "default_model" in data else None,
        "skills": data["skills"] if "skills" in data else None,
        "skill_selection_mode": data["skill_selection_mode"]
        if "skill_selection_mode" in data
        else None,
        "enabled_skills": data["enabled_skills"] if "enabled_skills" in data else [],
        "disabled_providers": data["disabled_providers"]
        if "disabled_providers" in data
        else [],
        "api_key_providers": api_key_providers,
        "tool_permissions": data["tool_permissions"]
        if "tool_permissions" in data
        else {},
        "tool_enabled": data["tool_enabled"] if "tool_enabled" in data else {},
        "tool_max_calls": data["tool_max_calls"] if "tool_max_calls" in data else {},
        "compress_threshold": data["compress_threshold"]
        if "compress_threshold" in data
        else None,
        "reasoning_level": data["reasoning_level"] if "reasoning_level" in data else None,
    }


def _append_config_audit_entry(
    target: Path,
    *,
    before: ConfigAuditState | None,
    after: ConfigAuditState,
) -> None:
    audit_path = config_audit_path()
    _reject_test_write_to_real_home_audit(audit_path)
    audit_path.parent.mkdir(parents=True, exist_ok=True)
    payload = {
        "timestamp": datetime.now(timezone.utc).isoformat(),
        "actor": os.environ["USER"] if "USER" in os.environ else "",
        "pid": os.getpid(),
        "cwd": str(Path.cwd()),
        "target": str(target),
        "caller": _config_audit_caller(),
        "changed_fields": _config_changed_fields(before, after),
        "before": before,
        "after": after,
    }
    with _config_file_lock(audit_path, exclusive=True):
        with audit_path.open("a", encoding="utf-8") as stream:
            stream.write(json.dumps(payload, sort_keys=True) + "\n")
    audit_path.chmod(0o600)


def _config_changed_fields(
    before: ConfigAuditState | None,
    after: ConfigAuditState,
) -> list[str]:
    if before is None:
        return sorted(after.keys())
    return sorted(field for field in after if before[field] != after[field])


def _reject_test_write_to_real_home_audit(audit_path: Path) -> None:
    if "PYTEST_CURRENT_TEST" not in os.environ:
        return
    real_audit_path = _PROCESS_START_HOME / ".aceai" / "config.audit.jsonl"
    if audit_path.resolve() != real_audit_path.resolve():
        return
    raise RuntimeError(
        "Tests must not write the real AceAI config audit; isolate HOME with tmp_path"
    )


def _config_audit_entry_from_mapping(data: object) -> ConfigAuditEntry:
    if not isinstance(data, dict):
        raise TypeError("AceAI config audit entry must be a mapping")
    caller = data["caller"]
    if type(caller) is not list:
        raise TypeError("AceAI config audit caller must be a list")
    changed_fields = data["changed_fields"]
    if type(changed_fields) is not list:
        raise TypeError("AceAI config audit changed_fields must be a list")
    before = data["before"]
    after = data["after"]
    if before is not None and not isinstance(before, dict):
        raise TypeError("AceAI config audit before must be a mapping")
    if not isinstance(after, dict):
        raise TypeError("AceAI config audit after must be a mapping")
    return ConfigAuditEntry(
        timestamp=data["timestamp"],
        actor=data["actor"],
        pid=data["pid"],
        cwd=data["cwd"],
        target=data["target"],
        caller=tuple(caller),
        changed_fields=tuple(changed_fields),
        before=before,
        after=after,
    )


def _config_audit_caller() -> list[str]:
    callers: list[str] = []
    for frame in inspect.stack()[2:8]:
        callers.append(f"{Path(frame.filename).name}:{frame.lineno}:{frame.function}")
    return callers


def _load_api_keys(data: dict[object, object]) -> dict[str, str]:
    api_keys: dict[str, str] = {}
    if "api_keys" not in data:
        return api_keys
    raw_api_keys = data["api_keys"]
    if not isinstance(raw_api_keys, dict):
        raise TypeError("AceAI config api_keys must be a mapping")
    for provider, api_key in raw_api_keys.items():
        if type(provider) is not str:
            raise TypeError("AceAI config api_keys provider names must be str")
        if provider not in supported_provider_names():
            raise ValueError("AceAI config api_keys provider is unsupported")
        if type(api_key) is not str:
            raise TypeError("AceAI config api_keys values must be str")
        api_keys[provider] = api_key
    return api_keys


def _load_disabled_providers(data: dict[object, object]) -> list[str]:
    if "disabled_providers" not in data:
        return []
    raw_disabled_providers = data["disabled_providers"]
    if type(raw_disabled_providers) is not list:
        raise TypeError("AceAI config disabled_providers must be a list")
    disabled_providers: list[str] = []
    for provider in raw_disabled_providers:
        if type(provider) is not str:
            raise TypeError("AceAI config disabled_providers values must be str")
        if provider not in supported_provider_names():
            raise ValueError("AceAI config disabled_providers value is unsupported")
        disabled_providers.append(provider)
    return disabled_providers


def _select_loadable_provider(
    provider: str,
    api_keys: dict[str, str],
    disabled_providers: list[str],
) -> str | None:
    if provider in disabled_providers:
        return _first_loadable_provider(api_keys, disabled_providers)
    if _api_key_for_loadable_provider(provider, api_keys) != "":
        return provider
    return _first_loadable_provider(api_keys, disabled_providers)


def _first_loadable_provider(
    api_keys: dict[str, str],
    disabled_providers: list[str],
) -> str | None:
    for candidate in api_keys:
        if candidate in disabled_providers:
            continue
        if _api_key_for_loadable_provider(candidate, api_keys) != "":
            return candidate
    return None


def _api_key_for_loadable_provider(provider: str, api_keys: dict[str, str]) -> str:
    if provider in api_keys and api_keys[provider] != "":
        return api_keys[provider]
    env_name = api_key_env(provider)
    if env_name in os.environ:
        return os.environ[env_name]
    return default_api_key_for_provider(provider)


def _load_enabled_skills(data: dict[object, object]) -> list[str]:
    if "enabled_skills" not in data:
        return []
    raw_enabled_skills = data["enabled_skills"]
    if type(raw_enabled_skills) is not list:
        raise TypeError("AceAI config enabled_skills must be a list")
    enabled_skills: list[str] = []
    for skill_name in raw_enabled_skills:
        if type(skill_name) is not str:
            raise TypeError("AceAI config enabled_skills values must be str")
        enabled_skills.append(skill_name)
    return enabled_skills


def _load_tool_permissions(data: dict[object, object]) -> dict[str, ToolPermission]:
    if "tool_permissions" not in data:
        return {}
    raw_permissions = data["tool_permissions"]
    if not isinstance(raw_permissions, dict):
        raise TypeError("AceAI config tool_permissions must be a mapping")
    permissions: dict[str, ToolPermission] = {}
    for tool_name, permission in raw_permissions.items():
        if type(tool_name) is not str:
            raise TypeError("AceAI config tool_permissions tool names must be str")
        if permission not in ("always", "ask"):
            raise ValueError("AceAI config tool_permissions value is unsupported")
        permissions[tool_name] = permission
    return permissions


def _load_tool_enabled(data: dict[object, object]) -> dict[str, bool]:
    if "tool_enabled" not in data:
        return {}
    raw_enabled = data["tool_enabled"]
    if not isinstance(raw_enabled, dict):
        raise TypeError("AceAI config tool_enabled must be a mapping")
    enabled_by_tool: dict[str, bool] = {}
    for tool_name, enabled in raw_enabled.items():
        if type(tool_name) is not str:
            raise TypeError("AceAI config tool_enabled tool names must be str")
        if type(enabled) is not bool:
            raise TypeError("AceAI config tool_enabled values must be bool")
        enabled_by_tool[tool_name] = enabled
    return enabled_by_tool


def _load_tool_max_calls(data: dict[object, object]) -> dict[str, int]:
    if "tool_max_calls" not in data:
        return {}
    raw_max_calls = data["tool_max_calls"]
    if not isinstance(raw_max_calls, dict):
        raise TypeError("AceAI config tool_max_calls must be a mapping")
    max_calls_by_tool: dict[str, int] = {}
    for tool_name, max_calls in raw_max_calls.items():
        if type(tool_name) is not str:
            raise TypeError("AceAI config tool_max_calls tool names must be str")
        if type(max_calls) is not int:
            raise TypeError("AceAI config tool_max_calls values must be int")
        if max_calls < 1:
            raise ValueError("AceAI config tool_max_calls values must be positive")
        max_calls_by_tool[tool_name] = max_calls
    return max_calls_by_tool


def _load_compress_threshold(data: dict[object, object]) -> CompressThreshold:
    if "compress_threshold" not in data:
        return "100%"
    threshold = data["compress_threshold"]
    if type(threshold) is str:
        ContextCompressionPolicy(threshold)
        return threshold
    if type(threshold) is int:
        ContextCompressionPolicy(threshold)
        return threshold
    if type(threshold) is float:
        ContextCompressionPolicy(threshold)
        return threshold
    raise TypeError("AceAI config compress_threshold must be str, int, or float")


def _load_reasoning_level(data: dict[object, object]) -> ReasoningLevel:
    if "reasoning_level" not in data:
        return "auto"
    level = data["reasoning_level"]
    if type(level) is not str:
        raise TypeError("AceAI config reasoning_level must be str")
    if level not in ("auto", "low", "medium", "high", "max"):
        raise ValueError("AceAI config reasoning_level is unsupported")
    return level
